# Copyright 2026 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os

os.environ['JAX_TRACEBACK_FILTERING'] = 'off'

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from brainevent._csr.slice import csr_slice_rows, csr_slice_rows_p
from brainevent._csr.main import CSR, CSC
from brainevent._csr.test_util import get_csr

platform = jax.default_backend()
SLICE_IMPLEMENTATIONS = tuple(csr_slice_rows_p.available_backends(platform))


def _make_csr_and_dense(m, n, prob=0.3):
    """Create a CSR matrix and its dense equivalent for testing."""
    indptr, indices = get_csr(m, n, prob)
    data = jnp.asarray(np.random.randn(indices.shape[0]).astype(np.float32))
    dense = np.zeros((m, n), dtype=np.float32)
    indptr_np = np.asarray(indptr)
    indices_np = np.asarray(indices)
    data_np = np.asarray(data)
    for i in range(m):
        for j in range(indptr_np[i], indptr_np[i + 1]):
            dense[i, indices_np[j]] += data_np[j]
    return data, indices, indptr, jnp.asarray(dense)


@pytest.mark.skipif(
    not SLICE_IMPLEMENTATIONS,
    reason=f'No csr_slice_rows implementation on platform={platform}',
)
class TestCSRSliceRows:

    @pytest.mark.parametrize('implementation', SLICE_IMPLEMENTATIONS)
    def test_single_row(self, implementation):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        for row in [0, 3, m - 1]:
            row_indices = jnp.array([row], dtype=jnp.int32)
            result = csr_slice_rows(data, indices, indptr, row_indices,
                                    shape=(m, n), backend=implementation)
            expected = dense[row:row + 1]
            assert jnp.allclose(result, expected, atol=1e-5), \
                f"Row {row}: {result} != {expected}"

    @pytest.mark.parametrize('implementation', SLICE_IMPLEMENTATIONS)
    def test_multi_row(self, implementation):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        rows = [0, 3, 7]
        row_indices = jnp.array(rows, dtype=jnp.int32)
        result = csr_slice_rows(data, indices, indptr, row_indices,
                                shape=(m, n), backend=implementation)
        expected = dense[jnp.array(rows)]
        assert jnp.allclose(result, expected, atol=1e-5)

    @pytest.mark.parametrize('implementation', SLICE_IMPLEMENTATIONS)
    def test_array_index(self, implementation):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        row_indices = jnp.array([0, 2, 4, 6, 8], dtype=jnp.int32)
        result = csr_slice_rows(data, indices, indptr, row_indices,
                                shape=(m, n), backend=implementation)
        expected = dense[np.array([0, 2, 4, 6, 8])]
        assert jnp.allclose(result, expected, atol=1e-5)

    @pytest.mark.parametrize('implementation', SLICE_IMPLEMENTATIONS)
    def test_duplicate_rows(self, implementation):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        row_indices = jnp.array([2, 2, 5, 5], dtype=jnp.int32)
        result = csr_slice_rows(data, indices, indptr, row_indices,
                                shape=(m, n), backend=implementation)
        expected = dense[np.array([2, 2, 5, 5])]
        assert jnp.allclose(result, expected, atol=1e-5)

    @pytest.mark.parametrize('implementation', SLICE_IMPLEMENTATIONS)
    def test_oob_row(self, implementation):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        row_indices = jnp.array([999], dtype=jnp.int32)
        result = csr_slice_rows(data, indices, indptr, row_indices,
                                shape=(m, n), backend=implementation)
        expected = jnp.zeros((1, n), dtype=data.dtype)
        assert jnp.allclose(result, expected, atol=1e-5)

    @pytest.mark.parametrize('implementation', SLICE_IMPLEMENTATIONS)
    def test_all_rows(self, implementation):
        m, n = 8, 12
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        row_indices = jnp.arange(m, dtype=jnp.int32)
        result = csr_slice_rows(data, indices, indptr, row_indices,
                                shape=(m, n), backend=implementation)
        assert jnp.allclose(result, dense, atol=1e-5)


def _make_homo_csr_and_dense(m, n, prob=0.3):
    """Create a homogeneous-weight CSR matrix and its dense equivalent."""
    indptr, indices = get_csr(m, n, prob)
    data = jnp.array([1.5], dtype=jnp.float32)
    dense = np.zeros((m, n), dtype=np.float32)
    indptr_np = np.asarray(indptr)
    indices_np = np.asarray(indices)
    for i in range(m):
        for j in range(indptr_np[i], indptr_np[i + 1]):
            dense[i, indices_np[j]] += 1.5
    return data, indices, indptr, jnp.asarray(dense)


@pytest.mark.skipif(
    not SLICE_IMPLEMENTATIONS,
    reason=f'No csr_slice_rows implementation on platform={platform}',
)
class TestCSRSliceRowsHomo:
    """Tests for scalar (homogeneous) data support."""

    @pytest.mark.parametrize('implementation', SLICE_IMPLEMENTATIONS)
    def test_homo_single_row(self, implementation):
        m, n = 10, 15
        data, indices, indptr, dense = _make_homo_csr_and_dense(m, n)
        for row in [0, 3, m - 1]:
            row_indices = jnp.array([row], dtype=jnp.int32)
            result = csr_slice_rows(data, indices, indptr, row_indices, shape=(m, n), backend=implementation)
            expected = dense[row:row + 1]
            assert jnp.allclose(result, expected, atol=1e-5), f"Row {row}: {result} != {expected}"
            row = jnp.array(row, dtype=jnp.int32)
            result2 = csr_slice_rows(data, indices, indptr, row, shape=(m, n), backend=implementation)
            assert jnp.allclose(result2, expected[0], atol=1e-5)

    @pytest.mark.parametrize('implementation', SLICE_IMPLEMENTATIONS)
    def test_homo_multi_row(self, implementation):
        m, n = 10, 15
        data, indices, indptr, dense = _make_homo_csr_and_dense(m, n)
        rows = [0, 3, 7]
        row_indices = jnp.array(rows, dtype=jnp.int32)
        result = csr_slice_rows(data, indices, indptr, row_indices,
                                shape=(m, n), backend=implementation)
        expected = dense[jnp.array(rows)]
        assert jnp.allclose(result, expected, atol=1e-5)

    @pytest.mark.parametrize('implementation', SLICE_IMPLEMENTATIONS)
    def test_homo_all_rows(self, implementation):
        m, n = 8, 12
        data, indices, indptr, dense = _make_homo_csr_and_dense(m, n)
        row_indices = jnp.arange(m, dtype=jnp.int32)
        result = csr_slice_rows(data, indices, indptr, row_indices,
                                shape=(m, n), backend=implementation)
        assert jnp.allclose(result, dense, atol=1e-5)

    def test_homo_vjp(self):
        m, n = 10, 15
        data, indices, indptr, dense = _make_homo_csr_and_dense(m, n)
        row_indices = jnp.array([1, 3, 5], dtype=jnp.int32)

        def loss_fn(d):
            sliced = csr_slice_rows(d, indices, indptr, row_indices, shape=(m, n))
            return jnp.sum(sliced ** 2)

        grad_data = jax.grad(loss_fn)(data)
        assert grad_data.shape == (1,)

        eps = 1e-3
        fd = (loss_fn(data + eps) - loss_fn(data - eps)) / (2 * eps)
        assert abs(float(grad_data[0]) - float(fd)) < 1e-2, \
            f"Grad mismatch: analytic={grad_data[0]}, fd={fd}"

    def test_homo_jvp(self):
        m, n = 10, 15
        data, indices, indptr, dense = _make_homo_csr_and_dense(m, n)
        row_indices = jnp.array([0, 2, 4], dtype=jnp.int32)
        data_dot = jnp.ones_like(data)

        primals, tangents = jax.jvp(
            lambda d: csr_slice_rows(d, indices, indptr, row_indices, shape=(m, n)),
            (data,),
            (data_dot,),
        )
        expected_tangent = csr_slice_rows(data_dot, indices, indptr, row_indices, shape=(m, n))
        assert jnp.allclose(tangents, expected_tangent, atol=1e-5)


@pytest.mark.skipif(
    not SLICE_IMPLEMENTATIONS,
    reason=f'No csr_slice_rows implementation on platform={platform}',
)
class TestCSRGetitem:

    def test_single_int_index(self):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        csr = CSR(data, indices, indptr, shape=(m, n))
        result = csr[3]
        expected = dense[3]
        assert result.shape == (n,)
        assert jnp.allclose(result, expected, atol=1e-5)

    def test_tuple_index(self):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        csr = CSR(data, indices, indptr, shape=(m, n))
        result = csr[(0, 3, 7)]
        expected = dense[np.array([0, 3, 7])]
        assert result.shape == (3, n)
        assert jnp.allclose(result, expected, atol=1e-5)

    def test_list_index(self):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        csr = CSR(data, indices, indptr, shape=(m, n))
        result = csr[[1, 5, 9]]
        expected = dense[np.array([1, 5, 9])]
        assert result.shape == (3, n)
        assert jnp.allclose(result, expected, atol=1e-5)

    def test_array_index(self):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        csr = CSR(data, indices, indptr, shape=(m, n))
        idx = jnp.array([0, 2, 4], dtype=jnp.int32)
        result = csr[idx]
        expected = dense[np.array([0, 2, 4])]
        assert result.shape == (3, n)
        assert jnp.allclose(result, expected, atol=1e-5)

    def test_jit_single(self):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        csr = CSR(data, indices, indptr, shape=(m, n))
        result = jax.jit(lambda c: c[0])(csr)
        expected = dense[0]
        assert jnp.allclose(result, expected, atol=1e-5)

    def test_jit_multi(self):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        csr = CSR(data, indices, indptr, shape=(m, n))
        idx = jnp.array([1, 3, 5], dtype=jnp.int32)
        result = jax.jit(lambda c, i: c[i])(csr, idx)
        expected = dense[np.array([1, 3, 5])]
        assert jnp.allclose(result, expected, atol=1e-5)


@pytest.mark.skipif(
    not SLICE_IMPLEMENTATIONS,
    reason=f'No csr_slice_rows implementation on platform={platform}',
)
class TestCSRSliceAD:

    def test_jvp(self):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        row_indices = jnp.array([1, 3, 5], dtype=jnp.int32)
        data_dot = jnp.ones_like(data)

        primals, tangents = jax.jvp(
            lambda d: csr_slice_rows(d, indices, indptr, row_indices, shape=(m, n)),
            (data,),
            (data_dot,),
        )
        # Since the operation is linear in data, tangent should equal the
        # slice applied to data_dot
        expected_tangent = csr_slice_rows(data_dot, indices, indptr, row_indices, shape=(m, n))
        assert jnp.allclose(tangents, expected_tangent, atol=1e-5)

    def test_vjp(self):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        row_indices = jnp.array([1, 3, 5], dtype=jnp.int32)

        def loss_fn(d):
            sliced = csr_slice_rows(d, indices, indptr, row_indices, shape=(m, n))
            return jnp.sum(sliced ** 2)

        grad_data = jax.grad(loss_fn)(data)
        assert grad_data.shape == data.shape

        # Verify numerically with finite differences
        eps = 1e-3
        for idx in [0, data.shape[0] // 2, data.shape[0] - 1]:
            data_plus = data.at[idx].add(eps)
            data_minus = data.at[idx].add(-eps)
            fd = (loss_fn(data_plus) - loss_fn(data_minus)) / (2 * eps)
            assert abs(float(grad_data[idx]) - float(fd)) < 1e-2, \
                f"Grad mismatch at {idx}: analytic={grad_data[idx]}, fd={fd}"

    def test_grad_through_sum(self):
        m, n = 10, 15
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        row_indices = jnp.array([0, 5], dtype=jnp.int32)

        def loss_fn(d):
            return jnp.sum(csr_slice_rows(d, indices, indptr, row_indices, shape=(m, n)))

        grad_data = jax.grad(loss_fn)(data)
        assert grad_data.shape == data.shape


@pytest.mark.skipif(
    not SLICE_IMPLEMENTATIONS,
    reason=f'No csr_slice_rows implementation on platform={platform}',
)
class TestCSRSliceGradAD:
    """Tests for JVP and transpose rules on the gradient primitive itself."""

    def test_grad_of_grad(self):
        """Second-order gradient: grad(grad(loss)) should work via the grad primitive's transpose rule."""
        m, n = 8, 10
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        row_indices = jnp.array([1, 3], dtype=jnp.int32)

        def loss_fn(d):
            sliced = csr_slice_rows(d, indices, indptr, row_indices, shape=(m, n))
            return jnp.sum(sliced ** 3)

        # First-order gradient
        grad_fn = jax.grad(loss_fn)
        grad_data = grad_fn(data)
        assert grad_data.shape == data.shape

        # Second-order gradient (Hessian-vector product via grad-of-grad)
        def grad_sum(d):
            return jnp.sum(jax.grad(loss_fn)(d))

        hessian_diag_sum = jax.grad(grad_sum)(data)
        assert hessian_diag_sum.shape == data.shape

        # Verify numerically with finite differences on the gradient
        eps = 1e-3
        for idx in [0, data.shape[0] // 2]:
            data_plus = data.at[idx].add(eps)
            data_minus = data.at[idx].add(-eps)
            fd = (grad_fn(data_plus)[idx] - grad_fn(data_minus)[idx]) / (2 * eps)
            assert abs(float(hessian_diag_sum[idx]) - float(fd)) < 1e-1, \
                f"Second-order grad mismatch at {idx}: analytic={hessian_diag_sum[idx]}, fd={fd}"

    def test_jvp_through_grad(self):
        """JVP through the gradient primitive: jvp(grad(loss)) should work via the grad primitive's JVP rule."""
        m, n = 8, 10
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        row_indices = jnp.array([0, 2, 4], dtype=jnp.int32)

        def loss_fn(d):
            sliced = csr_slice_rows(d, indices, indptr, row_indices, shape=(m, n))
            return jnp.sum(sliced ** 2)

        grad_fn = jax.grad(loss_fn)
        data_dot = jnp.ones_like(data)

        # JVP through the gradient function
        primals_out, tangents_out = jax.jvp(grad_fn, (data,), (data_dot,))
        assert primals_out.shape == data.shape
        assert tangents_out.shape == data.shape

        # Verify numerically: tangent should approximate directional derivative of grad
        eps = 1e-4
        fd = (grad_fn(data + eps * data_dot) - grad_fn(data - eps * data_dot)) / (2 * eps)
        assert jnp.allclose(tangents_out, fd, atol=1e-2, rtol=1e-2), \
            f"JVP through grad mismatch: max diff={jnp.max(jnp.abs(tangents_out - fd))}"

    def test_grad_linear_loss(self):
        """For a linear loss (sum of sliced rows), grad is constant, so grad-of-grad should be zero."""
        m, n = 8, 10
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        row_indices = jnp.array([1, 5], dtype=jnp.int32)

        def linear_loss(d):
            return jnp.sum(csr_slice_rows(d, indices, indptr, row_indices, shape=(m, n)))

        # First-order gradient of a linear function is constant
        grad_data = jax.grad(linear_loss)(data)
        assert grad_data.shape == data.shape

        # Second-order gradient of a linear function is zero
        def grad_sum(d):
            return jnp.sum(jax.grad(linear_loss)(d))

        grad2 = jax.grad(grad_sum)(data)
        assert jnp.allclose(grad2, jnp.zeros_like(grad2), atol=1e-6)


@pytest.mark.skipif(
    not SLICE_IMPLEMENTATIONS,
    reason=f'No csr_slice_rows implementation on platform={platform}',
)
class TestCSRSliceBatching:
    """Tests for vmap (batching) through csr_slice_rows and its gradient."""

    def test_vmap_over_data(self):
        """vmap over data: batch of CSR matrices (same sparsity, different values)."""
        m, n = 8, 10
        _, indices, indptr, _ = _make_csr_and_dense(m, n)
        batch_size = 3
        batch_data = jnp.asarray(np.random.randn(batch_size, indices.shape[0]).astype(np.float32))
        row_indices = jnp.array([1, 3, 5], dtype=jnp.int32)

        result = jax.vmap(
            lambda d: csr_slice_rows(d, indices, indptr, row_indices, shape=(m, n))
        )(batch_data)
        assert result.shape == (batch_size, 3, n)

        for b in range(batch_size):
            expected = csr_slice_rows(batch_data[b], indices, indptr, row_indices, shape=(m, n))
            assert jnp.allclose(result[b], expected, atol=1e-5)

    def test_vmap_over_row_indices(self):
        """vmap over row_indices: different row subsets from the same CSR matrix."""
        m, n = 8, 10
        data, indices, indptr, dense = _make_csr_and_dense(m, n)
        batch_row_indices = jnp.array([[0, 1], [2, 3], [4, 5]], dtype=jnp.int32)

        result = jax.vmap(
            lambda ri: csr_slice_rows(data, indices, indptr, ri, shape=(m, n))
        )(batch_row_indices)
        assert result.shape == (3, 2, n)

        for b in range(3):
            expected = csr_slice_rows(data, indices, indptr, batch_row_indices[b], shape=(m, n))
            assert jnp.allclose(result[b], expected, atol=1e-5)

    def test_vmap_grad_over_data(self):
        """vmap(grad) over data: batched gradient computation."""
        m, n = 8, 10
        _, indices, indptr, _ = _make_csr_and_dense(m, n)
        batch_size = 3
        batch_data = jnp.asarray(np.random.randn(batch_size, indices.shape[0]).astype(np.float32))
        row_indices = jnp.array([1, 3], dtype=jnp.int32)

        def loss(d):
            return jnp.sum(csr_slice_rows(d, indices, indptr, row_indices, shape=(m, n)) ** 2)

        batched_grad = jax.vmap(jax.grad(loss))(batch_data)
        assert batched_grad.shape == batch_data.shape

        for b in range(batch_size):
            expected_grad = jax.grad(loss)(batch_data[b])
            assert jnp.allclose(batched_grad[b], expected_grad, atol=1e-5)

    def test_vmap_jit_combined(self):
        """vmap + jit combined."""
        m, n = 8, 10
        _, indices, indptr, _ = _make_csr_and_dense(m, n)
        batch_size = 4
        batch_data = jnp.asarray(np.random.randn(batch_size, indices.shape[0]).astype(np.float32))
        row_indices = jnp.array([0, 2, 4], dtype=jnp.int32)

        @jax.jit
        @jax.vmap
        def batched_slice(d):
            return csr_slice_rows(d, indices, indptr, row_indices, shape=(m, n))

        result = batched_slice(batch_data)
        assert result.shape == (batch_size, 3, n)

        for b in range(batch_size):
            expected = csr_slice_rows(batch_data[b], indices, indptr, row_indices, shape=(m, n))
            assert jnp.allclose(result[b], expected, atol=1e-5)


@pytest.mark.skipif(
    not SLICE_IMPLEMENTATIONS,
    reason=f'No csr_slice_rows implementation on platform={platform}',
)
class TestCSCGetitem:

    def test_single_col(self):
        m, n = 10, 15
        dense = np.random.randn(m, n).astype(np.float32)
        csc = CSC.fromdense(jnp.asarray(dense))
        result = csc[3]
        expected = dense[:, 3]
        assert result.shape == (m,)
        assert jnp.allclose(result, jnp.asarray(expected), atol=1e-5)

    def test_multi_col(self):
        m, n = 10, 15
        dense = np.random.randn(m, n).astype(np.float32)
        csc = CSC.fromdense(jnp.asarray(dense))
        result = csc[(0, 3, 7)]
        expected = dense[:, [0, 3, 7]]
        assert result.shape == (m, 3)
        assert jnp.allclose(result, jnp.asarray(expected), atol=1e-5)

    def test_array_col(self):
        m, n = 10, 15
        dense = np.random.randn(m, n).astype(np.float32)
        csc = CSC.fromdense(jnp.asarray(dense))
        idx = jnp.array([1, 5, 9], dtype=jnp.int32)
        result = csc[idx]
        expected = dense[:, [1, 5, 9]]
        assert result.shape == (m, 3)
        assert jnp.allclose(result, jnp.asarray(expected), atol=1e-5)



# Copyright 2024 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numbers
from functools import partial
from typing import Union, Optional

import brainunit as u
import jax
import jax.numpy as jnp
import numpy as np

from brainevent._misc import namescope
from brainevent._op import XLACustomKernel, numba_kernel, jaxinfo_to_warpinfo
from brainevent._op.benchmark import BenchmarkConfig
from brainevent._typing import MatrixShape

__all__ = [
    'update_csr_on_binary_pre',
    'update_csr_on_binary_pre_p',
    'update_csr_on_binary_post',
    'update_csr_on_binary_post_p',
]


@namescope(static_argnames=['shape'])
def update_csr_on_binary_pre(
    weight: Union[u.Quantity, jax.Array, numbers.Number],
    indices: Union[np.ndarray, jax.Array],
    indptr: Union[np.ndarray, jax.Array],
    pre_spike: jax.Array,
    post_trace: Union[u.Quantity, jax.Array],
    w_min: Optional[Union[u.Quantity, jax.Array, numbers.Number]] = None,
    w_max: Optional[Union[u.Quantity, jax.Array, numbers.Number]] = None,
    *,
    shape: MatrixShape,
    backend: Optional[str] = None,
):
    """Update CSR synaptic weights triggered by presynaptic binary spike events.

    Implements a spike-timing-dependent plasticity (STDP) rule for sparse
    connectivity matrices stored in Compressed Sparse Row (CSR) format. For each
    presynaptic neuron ``i`` that fires (``pre_spike[i]`` is ``True`` or nonzero),
    the weights of all outgoing synapses from that neuron are updated by adding the
    corresponding postsynaptic trace values:

    ``weight[indptr[i]:indptr[i+1]] += post_trace[indices[indptr[i]:indptr[i+1]]]``

    After the update, weights are optionally clipped to ``[w_min, w_max]``.

    Parameters
    ----------
    weight : jax.Array, Quantity, or number
        Sparse synaptic weight array in CSR data format, with shape ``(nse,)``
        where ``nse`` is the number of stored elements. May carry physical units
        via ``brainunit.Quantity``.
    indices : ndarray or jax.Array
        Column indices array of the CSR format, with shape ``(nse,)`` and integer
        dtype.
    indptr : ndarray or jax.Array
        Row pointer array of the CSR format, with shape ``(n_pre + 1,)`` and
        integer dtype.
    pre_spike : jax.Array
        Binary or boolean array indicating presynaptic spike events, with shape
        ``(n_pre,)``. If boolean, ``True`` indicates a spike. If float, any
        nonzero value indicates a spike.
    post_trace : jax.Array or Quantity
        Postsynaptic eligibility trace values, with shape ``(n_post,)``. Must be
        compatible in units with ``weight``.
    w_min : jax.Array, Quantity, number, or None, optional
        Lower bound for weight clipping. Must have the same units as ``weight``.
        If ``None``, no lower bound is applied.
    w_max : jax.Array, Quantity, number, or None, optional
        Upper bound for weight clipping. Must have the same units as ``weight``.
        If ``None``, no upper bound is applied.
    shape : tuple of int
        Full matrix shape as ``(n_pre, n_post)``.
    backend : str or None, optional
        Compute backend to use. One of ``'numba'``, ``'warp'``, ``'pallas'``, or
        ``None`` for automatic selection.

    Returns
    -------
    jax.Array or Quantity
        Updated weight array with the same shape ``(nse,)`` and units as the
        input ``weight``.

    Raises
    ------
    AssertionError
        If ``weight`` is not 1-D, if ``pre_spike`` is not 1-D, if ``post_trace``
        is not 1-D, if ``shape[0] != pre_spike.shape[0]``,
        ``shape[1] != post_trace.shape[0]``, or if
        ``weight.shape[0] != indices.shape[0]``.  These checks are performed by
        the underlying :func:`csr_on_pre_prim_call`.

    See Also
    --------
    update_csr_on_binary_post : Post-synaptic-spike-triggered weight update.
    update_csr_on_binary_pre_p : Low-level XLA custom kernel primitive for this
        operation.

    Notes
    -----
    This function implements the **pre-synaptic** component of an additive
    spike-timing-dependent plasticity (STDP) rule.  In the standard pair-based
    STDP formulation the weight matrix ``W`` with shape ``(n_pre, n_post)`` is
    stored in CSR format.  When presynaptic neuron ``j`` fires, the update for
    every synapse ``(i, j)`` that exists in the sparsity pattern is:

    ``W[i, j] <- W[i, j] + post_trace[i]``

    After the additive update, weights are clipped element-wise:

    ``W[i, j] <- clip(W[i, j], w_min, w_max)``

    Here ``post_trace`` is an eligibility trace that typically decays
    exponentially between postsynaptic spikes, so synapses that experienced a
    recent postsynaptic spike receive a larger update.

    The function internally converts ``post_trace`` to the same unit as ``weight``
    before performing arithmetic, so mixed-unit inputs are supported as long as
    the units are dimensionally compatible.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._csr.plasticity_binary import update_csr_on_binary_pre
        >>> weight = jnp.array([0.5, 0.3, 0.8, 0.2], dtype=jnp.float32)
        >>> indices = jnp.array([0, 1, 0, 2], dtype=jnp.int32)
        >>> indptr = jnp.array([0, 2, 4], dtype=jnp.int32)
        >>> pre_spike = jnp.array([True, False])
        >>> post_trace = jnp.array([0.1, 0.2, 0.05], dtype=jnp.float32)
        >>> updated = update_csr_on_binary_pre(
        ...     weight, indices, indptr, pre_spike, post_trace,
        ...     shape=(2, 3),
        ... )
    """
    weight, wunit = u.split_mantissa_unit(weight)
    post_trace = u.Quantity(post_trace).to(wunit).mantissa
    weight = u.maybe_decimal(
        csr_on_pre_prim_call(
            weight, indices, indptr, pre_spike, post_trace, shape=shape, backend=backend
        )[0] * wunit
    )
    weight = u.math.clip(weight, w_min, w_max)
    return weight


def _csr_on_pre_numba_kernel_generator(
    spike_info: jax.ShapeDtypeStruct,
    **kwargs
):
    import numba

    if spike_info.dtype == jnp.bool_:
        @numba.njit(fastmath=True)
        def kernel(weight, indices, indptr, pre_spike, post_trace, out_w):
            for i in range(pre_spike.shape[0]):
                if pre_spike[i]:
                    i_start = indptr[i]
                    i_end = indptr[i + 1]
                    out_w[i_start: i_end] += post_trace[indices[i_start: i_end]]
    else:
        @numba.njit(fastmath=True)
        def kernel(weight, indices, indptr, pre_spike, post_trace, out_w):
            for i in range(pre_spike.shape[0]):
                if pre_spike[i] != 0.:
                    i_start = indptr[i]
                    i_end = indptr[i + 1]
                    out_w[i_start: i_end] += post_trace[indices[i_start: i_end]]

    def fn(weight, indices, indptr, pre_spike, post_trace):
        return numba_kernel(kernel, outs=kwargs['outs'], input_output_aliases={0: 0})(
            weight, indices, indptr, pre_spike, post_trace
        )

    return fn


def _csr_on_pre_warp_kernel_generator(
    weight_info: jax.ShapeDtypeStruct,
    indices_info: jax.ShapeDtypeStruct,
    indptr_info: jax.ShapeDtypeStruct,
    spike_info: jax.ShapeDtypeStruct,
    trace_info: jax.ShapeDtypeStruct,
    shape: MatrixShape,
    **kwargs
):
    import warp
    from warp.jax_experimental import jax_kernel

    indices_warp_info = jaxinfo_to_warpinfo(indices_info)
    indptr_warp_info = jaxinfo_to_warpinfo(indptr_info)
    spike_warp_info = jaxinfo_to_warpinfo(spike_info)
    trace_warp_info = jaxinfo_to_warpinfo(trace_info)
    out_warp_info = jaxinfo_to_warpinfo(kwargs['outs'][0])

    if spike_info.dtype == jnp.bool_:
        @warp.kernel
        def plasticity_kernel(
            indices: indices_warp_info,
            indptr: indptr_warp_info,
            pre_spike: spike_warp_info,
            post_trace: trace_warp_info,
            out_w: out_warp_info,
        ):
            i = warp.tid()
            if pre_spike[i]:
                for j in range(indptr[i], indptr[i + 1]):
                    out_w[j] += post_trace[indices[j]]
    else:
        @warp.kernel
        def plasticity_kernel(
            indices: indices_warp_info,
            indptr: indptr_warp_info,
            pre_spike: spike_warp_info,
            post_trace: trace_warp_info,
            out_w: out_warp_info,
        ):
            i = warp.tid()
            if pre_spike[i] != 0.:
                for j in range(indptr[i], indptr[i + 1]):
                    out_w[j] += post_trace[indices[j]]

    def kernel(weight, indices, indptr, pre_spike, post_trace):
        fn = jax_kernel(plasticity_kernel, launch_dims=[shape[0]], num_outputs=1, in_out_argnames=['out_w'])
        return fn(indices, indptr, pre_spike, post_trace, weight)

    return kernel


def _csr_on_pre_pallas_kernel_generator(
    impl_backend,
    spike_info: jax.ShapeDtypeStruct,
    shape: MatrixShape,
    **kwargs
):
    from jax.experimental import pallas as pl

    if spike_info.dtype == jnp.bool_:
        def kernel_fn(weight_ref, indices_ref, indptr_ref, spike_ref, trace_ref, _, out_w_ref):
            i_row = pl.program_id(0)
            i_col_start = indptr_ref[i_row]
            i_col_end = indptr_ref[i_row + 1]

            @pl.when(spike_ref[i_row])
            def run():
                def loop_fn(j, _):
                    idx = i_col_start + j
                    post_id = indices_ref[idx]
                    out_w_ref[idx] = out_w_ref[idx] + trace_ref[post_id]

                jax.lax.fori_loop(0, i_col_end - i_col_start, loop_fn, None)
    else:
        def kernel_fn(weight_ref, indices_ref, indptr_ref, spike_ref, trace_ref, _, out_w_ref):
            i_row = pl.program_id(0)
            i_col_start = indptr_ref[i_row]
            i_col_end = indptr_ref[i_row + 1]

            @pl.when(spike_ref[i_row] != 0.)
            def run():
                def loop_fn(j, _):
                    idx = i_col_start + j
                    post_id = indices_ref[idx]
                    out_w_ref[idx] = out_w_ref[idx] + trace_ref[post_id]

                jax.lax.fori_loop(0, i_col_end - i_col_start, loop_fn, None)

    def kernel(weight, indices, indptr, pre_spike, post_trace):
        fn = pl.pallas_call(
            kernel_fn,
            grid=(shape[0],),
            input_output_aliases={5: 0},
            out_shape=kwargs['outs'],
            backend=impl_backend,
        )
        return fn(weight, indices, indptr, pre_spike, post_trace, weight)

    return kernel


def _csr_on_pre_benchmark_data(*, platform):
    n_pre, n_post, prob, dtype = 1000, 1000, 0.1, jnp.float32
    configs = []
    for bool_event in (True, False):
        n_conn = max(1, int(n_post * prob))
        indptr = np.arange(n_pre + 1, dtype=np.int32) * n_conn
        indices = np.random.randint(0, n_post, (n_pre * n_conn,), dtype=np.int32)
        weight = jnp.ones(n_pre * n_conn, dtype=dtype)
        if bool_event:
            pre_spike = jnp.asarray(np.random.rand(n_pre) > 0.5, dtype=jnp.bool_)
        else:
            pre_spike = jnp.asarray(np.random.rand(n_pre), dtype=dtype)
        post_trace = jnp.asarray(np.random.randn(n_post), dtype=dtype)
        name = f"{'bool' if bool_event else 'float'}"
        configs.append(
            BenchmarkConfig(
                name,
                (weight, indices, jnp.asarray(indptr), pre_spike, post_trace),
                {'shape': (n_pre, n_post)}
            )
        )
    return configs


def csr_on_pre_prim_call(weight, indices, indptr, pre_spike, post_trace, *, shape, backend: Optional[str] = None):
    """Invoke the low-level XLA custom kernel for presynaptic plasticity updates.

    Validates input shapes and dimensions, then dispatches to
    :data:`update_csr_on_binary_pre_p` with the appropriate metadata. This is the
    direct primitive call without unit handling or weight clipping; most users
    should prefer :func:`update_csr_on_binary_pre`.

    Parameters
    ----------
    weight : jax.Array
        Sparse synaptic weight data array, with shape ``(nse,)`` where ``nse`` is
        the number of stored elements.
    indices : jax.Array
        Column indices array of the CSR format, with shape ``(nse,)``.
    indptr : jax.Array
        Row pointer array of the CSR format, with shape ``(n_pre + 1,)``.
    pre_spike : jax.Array
        Binary or boolean presynaptic spike array, with shape ``(n_pre,)``.
    post_trace : jax.Array
        Postsynaptic trace values, with shape ``(n_post,)``.
    shape : tuple of int
        Full matrix shape as ``(n_pre, n_post)``.
    backend : str or None, optional
        Compute backend to use. One of ``'numba'``, ``'warp'``, ``'pallas'``, or
        ``None`` for automatic selection.

    Returns
    -------
    tuple of jax.Array
        A single-element tuple containing the updated weight array with shape
        ``(nse,)``.

    Raises
    ------
    AssertionError
        If ``weight`` is not 1-D, if ``pre_spike`` is not 1-D, if ``post_trace``
        is not 1-D, if dimension sizes in ``shape`` do not match
        ``pre_spike.shape[0]`` and ``post_trace.shape[0]``, or if
        ``weight.shape[0] != indices.shape[0]``.

    See Also
    --------
    update_csr_on_binary_pre : High-level wrapper with unit handling and clipping.

    Notes
    -----
    This function operates on unitless mantissa arrays. All physical-unit
    handling and weight clipping are performed by the caller
    (:func:`update_csr_on_binary_pre`).  The additive update rule applied per
    row ``i`` where ``pre_spike[i]`` is active is:

    ``weight[indptr[i] : indptr[i+1]] += post_trace[indices[indptr[i] : indptr[i+1]]]``

    The function constructs :class:`jax.ShapeDtypeStruct` metadata for each
    operand and forwards the call to the ``XLACustomKernel`` instance
    :data:`update_csr_on_binary_pre_p`.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._csr.plasticity_binary import csr_on_pre_prim_call
        >>> weight = jnp.array([0.5, 0.3, 0.8, 0.2], dtype=jnp.float32)
        >>> indices = jnp.array([0, 1, 0, 2], dtype=jnp.int32)
        >>> indptr = jnp.array([0, 2, 4], dtype=jnp.int32)
        >>> pre_spike = jnp.array([True, False])
        >>> post_trace = jnp.array([0.1, 0.2, 0.05], dtype=jnp.float32)
        >>> (updated,) = csr_on_pre_prim_call(
        ...     weight, indices, indptr, pre_spike, post_trace,
        ...     shape=(2, 3),
        ... )
    """
    assert weight.ndim == 1, 'dense_one_pre only support 1D weight.'
    assert pre_spike.ndim == 1, 'pre_spike should be 1D.'
    assert post_trace.ndim == 1, 'post_trace should be 1D.'
    assert shape[0] == pre_spike.shape[0], f'pre_spike shape {pre_spike.shape} does not match with shape {shape}.'
    assert shape[1] == post_trace.shape[0], f'post_trace shape {post_trace.shape} does not match with shape {shape}.'
    assert weight.shape[0] == indices.shape[0], (
        f'weight shape {weight.shape}, indices shape {indices.shape}, indptr shape {indptr.shape} do not match.'
    )
    return update_csr_on_binary_pre_p(
        weight, indices, indptr, pre_spike, post_trace,
        outs=[jax.ShapeDtypeStruct(weight.shape, weight.dtype)],
        shape=shape,
        weight_info=jax.ShapeDtypeStruct(weight.shape, weight.dtype),
        indices_info=jax.ShapeDtypeStruct(indices.shape, indices.dtype),
        indptr_info=jax.ShapeDtypeStruct(indptr.shape, indptr.dtype),
        spike_info=jax.ShapeDtypeStruct(pre_spike.shape, pre_spike.dtype),
        trace_info=jax.ShapeDtypeStruct(post_trace.shape, post_trace.dtype),
        backend=backend,
    )


update_csr_on_binary_pre_p = XLACustomKernel(
    'binary_csr_plast',
    doc="""
Low-level XLA custom-kernel primitive for ``update_csr_on_binary_pre``.

This ``XLACustomKernel`` instance dispatches the CSR weight update for pre-synaptic binary plasticity
operation to registered backends (``numba``, ``warp``, ``pallas``),
using runtime shape/dtype metadata provided by the high-level wrapper.

For each presynaptic neuron that fires, updates all outgoing synaptic weights
by adding the corresponding postsynaptic trace values, implementing the pre-synaptic
component of spike-timing-dependent plasticity (STDP).

Beyond backend dispatch, the primitive stores JAX transformation bindings
(JVP, transpose, batching, and call registration) so the operation integrates
correctly with ``jit``, ``vmap``, and autodiff.

Available backends can be queried with ``update_csr_on_binary_pre_p.available_backends(platform)``,
and the default backend can be configured with ``update_csr_on_binary_pre_p.set_default(platform, backend)``.

See Also
--------
update_csr_on_binary_pre : High-level user-facing function wrapper.
"""
)
update_csr_on_binary_pre_p.def_numba_kernel(_csr_on_pre_numba_kernel_generator)
update_csr_on_binary_pre_p.def_warp_kernel(_csr_on_pre_warp_kernel_generator)
update_csr_on_binary_pre_p.def_pallas_kernel('gpu', partial(_csr_on_pre_pallas_kernel_generator, 'triton'))
update_csr_on_binary_pre_p.def_pallas_kernel('tpu', partial(_csr_on_pre_pallas_kernel_generator, 'mosaic_tpu'))
update_csr_on_binary_pre_p.def_tags('csr', 'plasticity')
update_csr_on_binary_pre_p.def_benchmark_data(_csr_on_pre_benchmark_data)


@namescope(static_argnames=['shape'])
def update_csr_on_binary_post(
    weight: Union[u.Quantity, jax.Array],
    indices: Union[np.ndarray, jax.Array],
    indptr: Union[np.ndarray, jax.Array],
    weight_indices: Union[np.ndarray, jax.Array],
    pre_trace: Union[u.Quantity, jax.Array],
    post_spike: jax.Array,
    w_min: Optional[Union[u.Quantity, jax.Array, numbers.Number]] = None,
    w_max: Optional[Union[u.Quantity, jax.Array, numbers.Number]] = None,
    *,
    shape: MatrixShape,
    backend: Optional[str] = None,
):
    """Update CSR synaptic weights triggered by postsynaptic binary spike events.

    Implements a spike-timing-dependent plasticity (STDP) rule for sparse
    connectivity stored in CSC (Compressed Sparse Column) layout. For each
    postsynaptic neuron ``j`` that fires (``post_spike[j]`` is ``True`` or
    nonzero), the weights of all incoming synapses to that neuron are updated
    by adding the corresponding presynaptic trace values:

    ``weight[weight_indices[indptr[j]:indptr[j+1]]] += pre_trace[indices[indptr[j]:indptr[j+1]]]``

    The CSC structure (``indices``, ``indptr``) indexes by postsynaptic neuron,
    while ``weight_indices`` maps back to the original CSR weight positions.
    After the update, weights are optionally clipped to ``[w_min, w_max]``.

    Parameters
    ----------
    weight : jax.Array or Quantity
        Sparse synaptic weight array, with shape ``(nse,)`` where ``nse`` is the
        number of stored elements. May carry physical units via
        ``brainunit.Quantity``.
    indices : ndarray or jax.Array
        Row indices array of the CSC format, with shape ``(nse,)`` and integer
        dtype.
    indptr : ndarray or jax.Array
        Column pointer array of the CSC format, with shape ``(n_post + 1,)`` and
        integer dtype.
    weight_indices : ndarray or jax.Array
        Mapping from CSC element positions to CSR weight positions, with shape
        ``(nse,)`` and integer dtype.
    pre_trace : jax.Array or Quantity
        Presynaptic eligibility trace values, with shape ``(n_pre,)``. Must be
        compatible in units with ``weight``.
    post_spike : jax.Array
        Binary or boolean array indicating postsynaptic spike events, with shape
        ``(n_post,)``. If boolean, ``True`` indicates a spike. If float, any
        nonzero value indicates a spike.
    w_min : jax.Array, Quantity, number, or None, optional
        Lower bound for weight clipping. Must have the same units as ``weight``.
        If ``None``, no lower bound is applied.
    w_max : jax.Array, Quantity, number, or None, optional
        Upper bound for weight clipping. Must have the same units as ``weight``.
        If ``None``, no upper bound is applied.
    shape : tuple of int
        Full matrix shape as ``(n_pre, n_post)``.
    backend : str or None, optional
        Compute backend to use. One of ``'numba'``, ``'warp'``, ``'pallas'``, or
        ``None`` for automatic selection.

    Returns
    -------
    jax.Array or Quantity
        Updated weight array with the same shape ``(nse,)`` and units as the
        input ``weight``.

    Raises
    ------
    AssertionError
        If ``weight`` is not 1-D, if ``post_spike`` is not 1-D, if ``pre_trace``
        is not 1-D, if ``shape[1] != post_spike.shape[0]``,
        ``shape[0] != pre_trace.shape[0]``, or if ``weight.shape``,
        ``weight_indices.shape``, and ``indices.shape`` are not all equal.
        These checks are performed by the underlying
        :func:`csr2csc_on_post_prim_call`.

    See Also
    --------
    update_csr_on_binary_pre : Pre-synaptic-spike-triggered weight update.
    update_csr_on_binary_post_p : Low-level XLA custom kernel primitive for this
        operation.

    Notes
    -----
    This function implements the **post-synaptic** component of an additive
    spike-timing-dependent plasticity (STDP) rule.  In the standard pair-based
    STDP formulation the weight matrix ``W`` with shape ``(n_pre, n_post)`` is
    stored in CSR format.  When postsynaptic neuron ``i`` fires, the update for
    every synapse ``(i, j)`` that exists in the sparsity pattern is:

    ``W[i, j] <- W[i, j] + pre_trace[j]``

    After the additive update, weights are clipped element-wise:

    ``W[i, j] <- clip(W[i, j], w_min, w_max)``

    Here ``pre_trace`` is an eligibility trace that typically decays
    exponentially between presynaptic spikes, so synapses whose presynaptic
    neuron fired recently receive a larger update.

    The function internally converts ``pre_trace`` to the same unit as ``weight``
    before performing arithmetic, so mixed-unit inputs are supported as long as
    the units are dimensionally compatible.

    The CSC layout is used so that iterating over postsynaptic spikes efficiently
    gathers all incoming synapses. The ``weight_indices`` array allows writing
    the updated values back to the correct positions in the original CSR weight
    array.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._csr.plasticity_binary import update_csr_on_binary_post
        >>> weight = jnp.array([0.5, 0.3, 0.8, 0.2], dtype=jnp.float32)
        >>> indices = jnp.array([0, 1, 0, 1], dtype=jnp.int32)
        >>> indptr = jnp.array([0, 2, 4], dtype=jnp.int32)
        >>> weight_indices = jnp.array([0, 2, 1, 3], dtype=jnp.int32)
        >>> pre_trace = jnp.array([0.1, -0.05], dtype=jnp.float32)
        >>> post_spike = jnp.array([True, False])
        >>> updated = update_csr_on_binary_post(
        ...     weight, indices, indptr, weight_indices,
        ...     pre_trace, post_spike, shape=(2, 2),
        ... )
    """
    weight, wunit = u.split_mantissa_unit(weight)
    pre_trace = u.Quantity(pre_trace).to(wunit).mantissa
    weight = u.maybe_decimal(
        csr2csc_on_post_prim_call(
            weight, indices, indptr, weight_indices, pre_trace, post_spike, shape=shape, backend=backend
        )[0] * wunit
    )
    weight = u.math.clip(weight, w_min, w_max)
    return weight


def _csr2csc_on_post_numba_kernel_generator(
    spike_info: jax.ShapeDtypeStruct,
    **kwargs
):
    import numba

    # Note: Cannot parallelize due to potential race conditions when updating out_w[weight_ids]
    if spike_info.dtype == jnp.bool_:
        @numba.njit(fastmath=True)
        def kernel(weight, indices, indptr, weight_indices, pre_trace, post_spike, out_w):
            for i in range(post_spike.shape[0]):
                if post_spike[i]:
                    index = indptr[i]
                    index_end = indptr[i + 1]
                    weight_ids = weight_indices[index: index_end]
                    pre_ids = indices[index: index_end]
                    out_w[weight_ids] += pre_trace[pre_ids]
    else:
        @numba.njit(fastmath=True)
        def kernel(weight, indices, indptr, weight_indices, pre_trace, post_spike, out_w):
            for i in range(post_spike.shape[0]):
                if post_spike[i] != 0.:
                    index = indptr[i]
                    index_end = indptr[i + 1]
                    weight_ids = weight_indices[index: index_end]
                    pre_ids = indices[index: index_end]
                    out_w[weight_ids] += pre_trace[pre_ids]

    def fn(weight, indices, indptr, weight_indices, pre_trace, post_spike):
        return numba_kernel(kernel, outs=kwargs['outs'], input_output_aliases={0: 0})(
            weight, indices, indptr, weight_indices, pre_trace, post_spike
        )

    return fn


def _csr2csc_on_post_warp_kernel_generator(
    weight_info: jax.ShapeDtypeStruct,
    indices_info: jax.ShapeDtypeStruct,
    indptr_info: jax.ShapeDtypeStruct,
    weight_indices_info: jax.ShapeDtypeStruct,
    trace_info: jax.ShapeDtypeStruct,
    spike_info: jax.ShapeDtypeStruct,
    shape: MatrixShape,
    **kwargs
):
    import warp
    from warp.jax_experimental import jax_kernel

    indices_warp_info = jaxinfo_to_warpinfo(indices_info)
    indptr_warp_info = jaxinfo_to_warpinfo(indptr_info)
    weight_indices_warp_info = jaxinfo_to_warpinfo(weight_indices_info)
    trace_warp_info = jaxinfo_to_warpinfo(trace_info)
    spike_warp_info = jaxinfo_to_warpinfo(spike_info)
    out_warp_info = jaxinfo_to_warpinfo(kwargs['outs'][0])

    if spike_info.dtype == jnp.bool_:
        @warp.kernel
        def plasticity_kernel(
            indices: indices_warp_info,
            indptr: indptr_warp_info,
            weight_indices: weight_indices_warp_info,
            pre_trace: trace_warp_info,
            post_spike: spike_warp_info,
            out_w: out_warp_info,
        ):
            i = warp.tid()
            if post_spike[i]:
                for j in range(indptr[i], indptr[i + 1]):
                    weight_id = weight_indices[j]
                    pre_id = indices[j]
                    out_w[weight_id] += pre_trace[pre_id]
    else:
        @warp.kernel
        def plasticity_kernel(
            indices: indices_warp_info,
            indptr: indptr_warp_info,
            weight_indices: weight_indices_warp_info,
            pre_trace: trace_warp_info,
            post_spike: spike_warp_info,
            out_w: out_warp_info,
        ):
            i = warp.tid()
            if post_spike[i] != 0.:
                for j in range(indptr[i], indptr[i + 1]):
                    weight_id = weight_indices[j]
                    pre_id = indices[j]
                    out_w[weight_id] += pre_trace[pre_id]

    def kernel(weight, indices, indptr, weight_indices, pre_trace, post_spike):
        fn = jax_kernel(plasticity_kernel, launch_dims=[shape[1]], num_outputs=1, in_out_argnames=['out_w'])
        return fn(indices, indptr, weight_indices, pre_trace, post_spike, weight)

    return kernel


def _csr2csc_on_post_pallas_kernel_generator(
    impl_backend,
    spike_info: jax.ShapeDtypeStruct,
    shape: MatrixShape,
    **kwargs
):
    from jax.experimental import pallas as pl

    if spike_info.dtype == jnp.bool_:
        def kernel_fn(weight_ref, indices_ref, indptr_ref, weight_indices_ref, trace_ref, spike_ref, _, out_w_ref):
            i_col = pl.program_id(0)
            i_row_start = indptr_ref[i_col]
            i_row_end = indptr_ref[i_col + 1]

            @pl.when(spike_ref[i_col])
            def run():
                def loop_fn(j, _):
                    idx = i_row_start + j
                    pre_id = indices_ref[idx]
                    w_id = weight_indices_ref[idx]
                    out_w_ref[w_id] = out_w_ref[w_id] + trace_ref[pre_id]

                jax.lax.fori_loop(0, i_row_end - i_row_start, loop_fn, None)
    else:
        def kernel_fn(weight_ref, indices_ref, indptr_ref, weight_indices_ref, trace_ref, spike_ref, _, out_w_ref):
            i_col = pl.program_id(0)
            i_row_start = indptr_ref[i_col]
            i_row_end = indptr_ref[i_col + 1]

            @pl.when(spike_ref[i_col] != 0.)
            def run():
                def loop_fn(j, _):
                    idx = i_row_start + j
                    pre_id = indices_ref[idx]
                    w_id = weight_indices_ref[idx]
                    out_w_ref[w_id] = out_w_ref[w_id] + trace_ref[pre_id]

                jax.lax.fori_loop(0, i_row_end - i_row_start, loop_fn, None)

    def kernel(weight, indices, indptr, weight_indices, pre_trace, post_spike):
        fn = pl.pallas_call(
            kernel_fn, grid=(shape[1],), input_output_aliases={6: 0}, out_shape=kwargs['outs'], backend=impl_backend
        )
        return fn(weight, indices, indptr, weight_indices, pre_trace, post_spike, weight)

    return kernel


def _csr2csc_on_post_benchmark_data(*, platform):
    n_pre, n_post, prob, dtype = 1000, 1000, 0.1, jnp.float32
    configs = []
    for bool_event in (True, False):
        n_conn = max(1, int(n_post * prob))
        indptr = np.arange(n_post + 1, dtype=np.int32) * n_conn
        indices = np.random.randint(0, n_pre, (n_post * n_conn,), dtype=np.int32)
        weight = jnp.ones(n_post * n_conn, dtype=dtype)
        weight_indices = jnp.asarray(
            np.random.randint(0, n_post * n_conn, (n_post * n_conn,), dtype=np.int32)
        )
        pre_trace = jnp.asarray(np.random.randn(n_pre), dtype=dtype)
        if bool_event:
            post_spike = jnp.asarray(np.random.rand(n_post) > 0.5, dtype=jnp.bool_)
        else:
            post_spike = jnp.asarray(np.random.rand(n_post), dtype=dtype)
        name = f"{'bool' if bool_event else 'float'}"
        configs.append(
            BenchmarkConfig(
                name,
                (weight, indices, jnp.asarray(indptr), weight_indices, pre_trace, post_spike),
                {'shape': (n_pre, n_post)}
            )
        )
    return configs


def csr2csc_on_post_prim_call(weight, indices, indptr, weight_indices, pre_trace, post_spike, *, shape,
                              backend: Optional[str] = None):
    """Invoke the low-level XLA custom kernel for postsynaptic plasticity updates.

    Validates input shapes and dimensions, then dispatches to
    :data:`update_csr_on_binary_post_p` with the appropriate metadata. This is the
    direct primitive call without unit handling or weight clipping; most users
    should prefer :func:`update_csr_on_binary_post`.

    Parameters
    ----------
    weight : jax.Array
        Sparse synaptic weight data array, with shape ``(nse,)`` where ``nse`` is
        the number of stored elements.
    indices : jax.Array
        Row indices array of the CSC format, with shape ``(nse,)``.
    indptr : jax.Array
        Column pointer array of the CSC format, with shape ``(n_post + 1,)``.
    weight_indices : jax.Array
        Mapping from CSC element positions to CSR weight positions, with shape
        ``(nse,)`` and integer dtype.
    pre_trace : jax.Array
        Presynaptic trace values, with shape ``(n_pre,)``.
    post_spike : jax.Array
        Binary or boolean postsynaptic spike array, with shape ``(n_post,)``.
    shape : tuple of int
        Full matrix shape as ``(n_pre, n_post)``.
    backend : str or None, optional
        Compute backend to use. One of ``'numba'``, ``'warp'``, ``'pallas'``, or
        ``None`` for automatic selection.

    Returns
    -------
    tuple of jax.Array
        A single-element tuple containing the updated weight array with shape
        ``(nse,)``.

    Raises
    ------
    AssertionError
        If ``weight`` is not 1-D, if ``post_spike`` is not 1-D, if ``pre_trace``
        is not 1-D, if dimension sizes in ``shape`` do not match
        ``post_spike.shape[0]`` and ``pre_trace.shape[0]``, or if
        ``weight.shape``, ``weight_indices.shape``, and ``indices.shape`` are not
        all equal.

    See Also
    --------
    update_csr_on_binary_post : High-level wrapper with unit handling and clipping.

    Notes
    -----
    This function operates on unitless mantissa arrays. All physical-unit
    handling and weight clipping are performed by the caller
    (:func:`update_csr_on_binary_post`).  The CSC layout allows efficient
    iteration over postsynaptic neurons; for each postsynaptic neuron ``j``
    where ``post_spike[j]`` is active the update is:

    ``weight[weight_indices[indptr[j] : indptr[j+1]]] += pre_trace[indices[indptr[j] : indptr[j+1]]]``

    The function constructs :class:`jax.ShapeDtypeStruct` metadata for each
    operand and forwards the call to the ``XLACustomKernel`` instance
    :data:`update_csr_on_binary_post_p`.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._csr.plasticity_binary import csr2csc_on_post_prim_call
        >>> weight = jnp.array([0.5, 0.3, 0.8, 0.2], dtype=jnp.float32)
        >>> indices = jnp.array([0, 1, 0, 1], dtype=jnp.int32)
        >>> indptr = jnp.array([0, 2, 4], dtype=jnp.int32)
        >>> weight_indices = jnp.array([0, 2, 1, 3], dtype=jnp.int32)
        >>> pre_trace = jnp.array([0.1, -0.05], dtype=jnp.float32)
        >>> post_spike = jnp.array([True, False])
        >>> (updated,) = csr2csc_on_post_prim_call(
        ...     weight, indices, indptr, weight_indices,
        ...     pre_trace, post_spike, shape=(2, 2),
        ... )
    """
    assert weight.ndim == 1, 'dense_one_post only support 1D weight.'
    assert post_spike.ndim == 1, 'post_spike should be 1D.'
    assert pre_trace.ndim == 1, 'pre_trace should be 1D.'
    assert shape[1] == post_spike.shape[0], f'post_spike shape {post_spike.shape} does not match with shape {shape}.'
    assert shape[0] == pre_trace.shape[0], f'pre_trace shape {pre_trace.shape} does not match with shape {shape}.'
    assert weight.shape == weight_indices.shape == indices.shape, (
        f'weight shape {weight.shape}, weight_indices shape {weight_indices.shape}, '
        f'indices shape {indices.shape}, indptr shape {indptr.shape} do not match.'
    )
    return update_csr_on_binary_post_p(
        weight, indices, indptr, weight_indices, pre_trace, post_spike,
        outs=[jax.ShapeDtypeStruct(weight.shape, weight.dtype)],
        shape=shape,
        weight_info=jax.ShapeDtypeStruct(weight.shape, weight.dtype),
        indices_info=jax.ShapeDtypeStruct(indices.shape, indices.dtype),
        indptr_info=jax.ShapeDtypeStruct(indptr.shape, indptr.dtype),
        weight_indices_info=jax.ShapeDtypeStruct(weight_indices.shape, weight_indices.dtype),
        trace_info=jax.ShapeDtypeStruct(pre_trace.shape, pre_trace.dtype),
        spike_info=jax.ShapeDtypeStruct(post_spike.shape, post_spike.dtype),
        backend=backend,
    )


update_csr_on_binary_post_p = XLACustomKernel(
    'csr2csc_on_post',
    doc="""
Low-level XLA custom-kernel primitive for ``update_csr_on_binary_post``.

This ``XLACustomKernel`` instance dispatches the CSR weight update for post-synaptic binary plasticity (CSR to CSC conversion)
operation to registered backends (``numba``, ``warp``, ``pallas``),
using runtime shape/dtype metadata provided by the high-level wrapper.

For each postsynaptic neuron that fires, updates all incoming synaptic weights
by adding the corresponding presynaptic trace values, implementing the post-synaptic
component of spike-timing-dependent plasticity (STDP). Uses a CSC structure internally
for efficient iteration over postsynaptic spikes.

Beyond backend dispatch, the primitive stores JAX transformation bindings
(JVP, transpose, batching, and call registration) so the operation integrates
correctly with ``jit``, ``vmap``, and autodiff.

Available backends can be queried with ``update_csr_on_binary_post_p.available_backends(platform)``,
and the default backend can be configured with ``update_csr_on_binary_post_p.set_default(platform, backend)``.

See Also
--------
update_csr_on_binary_post : High-level user-facing function wrapper.
"""
)
update_csr_on_binary_post_p.def_numba_kernel(_csr2csc_on_post_numba_kernel_generator)
update_csr_on_binary_post_p.def_warp_kernel(_csr2csc_on_post_warp_kernel_generator)
update_csr_on_binary_post_p.def_pallas_kernel('gpu', partial(_csr2csc_on_post_pallas_kernel_generator, 'triton'))
update_csr_on_binary_post_p.def_pallas_kernel('tpu', partial(_csr2csc_on_post_pallas_kernel_generator, 'mosaic_tpu'))
update_csr_on_binary_post_p.def_tags('csr', 'plasticity')
update_csr_on_binary_post_p.def_benchmark_data(_csr2csc_on_post_benchmark_data)



# Copyright 2026 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from brainevent._coo.binary import binary_coomv, binary_coomm
from brainevent._coo.float import coomv, coomm, coomv_p, coomm_p

if jax.default_backend() != 'gpu':
    pytest.skip('Pallas GPU COO tests require a GPU backend.', allow_module_level=True)
if 'pallas' not in coomv_p.available_backends('gpu'):
    pytest.skip('Pallas backend is unavailable for coomv on GPU.', allow_module_level=True)
if 'pallas' not in coomm_p.available_backends('gpu'):
    pytest.skip('Pallas backend is unavailable for coomm on GPU.', allow_module_level=True)


def _make_indices(shape, nnz, mode, seed):
    m, n = shape
    rng = np.random.default_rng(seed)
    if nnz == 0:
        return jnp.empty((0,), dtype=jnp.int32), jnp.empty((0,), dtype=jnp.int32)

    if mode == 'uniform':
        row = rng.integers(0, m, size=nnz, dtype=np.int32)
        col = rng.integers(0, n, size=nnz, dtype=np.int32)
    elif mode == 'skewed_row':
        # Heavily concentrated rows to stress atomic contention and duplicate accumulation.
        hot = max(1, m // 16)
        hot_rows = rng.integers(0, hot, size=nnz, dtype=np.int32)
        cold_rows = rng.integers(0, m, size=nnz, dtype=np.int32)
        choose_hot = rng.random(nnz) < 0.85
        row = np.where(choose_hot, hot_rows, cold_rows).astype(np.int32)
        col = rng.integers(0, n, size=nnz, dtype=np.int32)
    else:
        raise ValueError(f'Unsupported mode: {mode}')

    return jnp.asarray(row), jnp.asarray(col)


def _expand_weights(weights, nnz):
    return jnp.full((nnz,), weights[0], dtype=weights.dtype) if weights.shape[0] == 1 else weights


def _dense_from_coo(weights, row, col, shape):
    dense = jnp.zeros(shape, dtype=weights.dtype)
    if row.shape[0] == 0:
        return dense
    return dense.at[(row, col)].add(_expand_weights(weights, row.shape[0]))


def _tol(dtype):
    if dtype == jnp.float32:
        return (1e-4, 1e-4)
    if dtype == jnp.float16:
        return (7e-2, 7e-2)
    return (3e-1, 3e-1)


def _tol_mm(dtype):
    # COO-MM accumulates many more atomic contributions than COO-MV.
    if dtype == jnp.float32:
        return (2e-2, 2e-2)
    return (8e-2, 8e-2)


@pytest.mark.parametrize('shape,nnz', [((1, 1), 0), ((17, 29), 129), ((257, 193), 2048)])
@pytest.mark.parametrize('mode', ['uniform', 'skewed_row'])
@pytest.mark.parametrize('transpose', [False, True])
@pytest.mark.parametrize('homo_w', [False, True])
@pytest.mark.parametrize('dtype', [jnp.float32, jnp.float16, jnp.bfloat16])
def test_pallas_coomv_matches_dense(shape, nnz, mode, transpose, homo_w, dtype):
    row, col = _make_indices(shape, nnz, mode, seed=11)
    rng = np.random.default_rng(12)

    if homo_w:
        weights = jnp.asarray([rng.standard_normal()], dtype=dtype)
    else:
        weights = jnp.asarray(rng.standard_normal(nnz), dtype=dtype)

    v_len = shape[0] if transpose else shape[1]
    v = jnp.asarray(rng.standard_normal(v_len), dtype=dtype)

    out = coomv(weights, row, col, v, shape=shape, transpose=transpose, backend='pallas')

    dense = _dense_from_coo(weights, row, col, shape)
    ref = dense.T @ v if transpose else dense @ v

    rtol, atol = _tol(dtype)
    assert jnp.allclose(out, ref, rtol=rtol, atol=atol)


@pytest.mark.parametrize('shape,nnz', [((1, 1), 0), ((17, 29), 129), ((257, 193), 2048)])
@pytest.mark.parametrize('mode', ['uniform', 'skewed_row'])
@pytest.mark.parametrize('transpose', [False, True])
@pytest.mark.parametrize('homo_w', [False, True])
@pytest.mark.parametrize('dtype', [jnp.float32, jnp.float16])
def test_pallas_coomm_matches_dense(shape, nnz, mode, transpose, homo_w, dtype):
    row, col = _make_indices(shape, nnz, mode, seed=21)
    rng = np.random.default_rng(22)

    if homo_w:
        weights = jnp.asarray([rng.standard_normal()], dtype=dtype)
    else:
        weights = jnp.asarray(rng.standard_normal(nnz), dtype=dtype)

    n_rhs = 32
    b_rows = shape[0] if transpose else shape[1]
    B = jnp.asarray(rng.standard_normal((b_rows, n_rhs)), dtype=dtype)

    out = coomm(weights, row, col, B, shape=shape, transpose=transpose, backend='pallas')

    dense = _dense_from_coo(weights, row, col, shape)
    ref = dense.T @ B if transpose else dense @ B

    rtol, atol = _tol_mm(dtype)
    assert jnp.allclose(out, ref, rtol=rtol, atol=atol)


@pytest.mark.parametrize('transpose', [False, True])
def test_pallas_binary_coo_matches_dense(transpose):
    shape = (33, 41)
    nnz = 511
    row, col = _make_indices(shape, nnz, mode='skewed_row', seed=31)
    rng = np.random.default_rng(32)
    weights = jnp.asarray(rng.standard_normal(nnz), dtype=jnp.float32)

    v_len = shape[0] if transpose else shape[1]
    bool_v = jnp.asarray(rng.random(v_len) > 0.7)
    float_v = jnp.asarray(rng.standard_normal(v_len), dtype=jnp.float32)

    dense = _dense_from_coo(weights, row, col, shape)

    out_bool = binary_coomv(weights, row, col, bool_v, shape=shape, transpose=transpose, backend='pallas')
    ref_bool = dense.T @ bool_v.astype(weights.dtype) if transpose else dense @ bool_v.astype(weights.dtype)
    assert jnp.allclose(out_bool, ref_bool, rtol=1e-5, atol=1e-5)

    out_float = binary_coomv(weights, row, col, float_v, shape=shape, transpose=transpose, backend='pallas')
    act = (float_v > 0).astype(weights.dtype)
    ref_float = dense.T @ act if transpose else dense @ act
    assert jnp.allclose(out_float, ref_float, rtol=1e-5, atol=1e-5)

    n_rhs = 24
    b_rows = shape[0] if transpose else shape[1]
    bool_B = jnp.asarray(rng.random((b_rows, n_rhs)) > 0.75)
    float_B = jnp.asarray(rng.standard_normal((b_rows, n_rhs)), dtype=jnp.float32)

    out_bool_mm = binary_coomm(weights, row, col, bool_B, shape=shape, transpose=transpose, backend='pallas')
    ref_bool_mm = dense.T @ bool_B.astype(weights.dtype) if transpose else dense @ bool_B.astype(weights.dtype)
    assert jnp.allclose(out_bool_mm, ref_bool_mm, rtol=5e-3, atol=5e-3)

    out_float_mm = binary_coomm(weights, row, col, float_B, shape=shape, transpose=transpose, backend='pallas')
    act_B = (float_B > 0).astype(weights.dtype)
    ref_float_mm = dense.T @ act_B if transpose else dense @ act_B
    assert jnp.allclose(out_float_mm, ref_float_mm, rtol=5e-3, atol=5e-3)


def test_pallas_empty_coo_outputs_zero():
    shape = (19, 23)
    row = jnp.empty((0,), dtype=jnp.int32)
    col = jnp.empty((0,), dtype=jnp.int32)
    weights = jnp.empty((0,), dtype=jnp.float32)

    v = jnp.ones((shape[1],), dtype=jnp.float32)
    B = jnp.ones((shape[1], 7), dtype=jnp.float32)

    out_mv = coomv(weights, row, col, v, shape=shape, transpose=False, backend='pallas')
    out_mm = coomm(weights, row, col, B, shape=shape, transpose=False, backend='pallas')

    assert out_mv.shape == (shape[0],)
    assert out_mm.shape == (shape[0], 7)
    assert jnp.all(out_mv == 0)
    assert jnp.all(out_mm == 0)


def test_pallas_determinism_with_atomics_tolerance():
    shape = (257, 193)
    nnz = 8192
    row, col = _make_indices(shape, nnz, mode='skewed_row', seed=41)
    rng = np.random.default_rng(42)

    weights = jnp.asarray(rng.standard_normal(nnz), dtype=jnp.float32)
    v = jnp.asarray(rng.standard_normal(shape[1]), dtype=jnp.float32)

    f = jax.jit(lambda w, r, c, x: coomv(w, r, c, x, shape=shape, transpose=False, backend='pallas'))
    y0 = f(weights, row, col, v)
    y1 = f(weights, row, col, v)
    y2 = f(weights, row, col, v)

    assert jnp.allclose(y0, y1, rtol=1e-4, atol=1e-4)
    assert jnp.allclose(y1, y2, rtol=1e-4, atol=1e-4)


def test_pallas_input_validation_errors():
    weights = jnp.asarray([1.0], dtype=jnp.float32)

    with pytest.raises(ValueError, match='same length'):
        coomv(weights, jnp.array([0, 1], dtype=jnp.int32), jnp.array([0], dtype=jnp.int32),
              jnp.ones((2,), dtype=jnp.float32),
              shape=(2, 2), transpose=False, backend='pallas')

    with pytest.raises(ValueError, match='length must be 1 or nnz'):
        coomv(jnp.array([1.0, 2.0], dtype=jnp.float32), jnp.array([0], dtype=jnp.int32),
              jnp.array([1], dtype=jnp.int32),
              jnp.ones((2,), dtype=jnp.float32), shape=(2, 2), transpose=False, backend='pallas')

    with pytest.raises(ValueError, match='incompatible shape'):
        coomm(weights, jnp.array([0], dtype=jnp.int32), jnp.array([1], dtype=jnp.int32),
              jnp.ones((3, 2), dtype=jnp.float32),
              shape=(2, 2), transpose=False, backend='pallas')


@pytest.mark.skipif(not jax.config.read('jax_enable_x64'), reason='Requires JAX x64 enabled for true int64 indices.')
def test_pallas_int64_indices_parity():
    shape = (41, 37)
    nnz = 733
    row, col = _make_indices(shape, nnz, mode='uniform', seed=51)
    row = row.astype(jnp.int64)
    col = col.astype(jnp.int64)

    rng = np.random.default_rng(52)
    weights = jnp.asarray(rng.standard_normal(nnz), dtype=jnp.float32)
    v = jnp.asarray(rng.standard_normal(shape[1]), dtype=jnp.float32)
    B = jnp.asarray(rng.standard_normal((shape[1], 8)), dtype=jnp.float32)

    out_mv = coomv(weights, row, col, v, shape=shape, transpose=False, backend='pallas')
    out_mm = coomm(weights, row, col, B, shape=shape, transpose=False, backend='pallas')

    dense = _dense_from_coo(weights, row, col, shape)
    ref_mv = dense @ v
    ref_mm = dense @ B

    assert jnp.allclose(out_mv, ref_mv, rtol=1e-4, atol=1e-4)
    assert jnp.allclose(out_mm, ref_mm, rtol=1e-4, atol=1e-4)



# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# -*- coding: utf-8 -*-

from typing import Optional

import brainunit as u
import jax
import numpy as np
from jax import numpy as jnp
from jax.interpreters import ad

from brainevent._compatible_import import Tracer
from brainevent._data import _initialize_seed, _initialize_conn_length
from brainevent._misc import generate_block_dim
from brainevent._numba_random import get_numba_lfsr_seed, get_numba_lfsr_random_integers, get_numba_lfsr_normal
from brainevent._op import XLACustomKernel, jaxinfo_to_warpinfo, numba_kernel, general_batching_rule
from brainevent._op.benchmark import BenchmarkConfig
from brainevent._pallas_random import get_pallas_lfsr_rng_class
from brainevent._typing import Data, MatrixShape

__all__ = [
    "jitn",
    "jitn_p",
    "jitnmv",
    "jitnmv_p",
    "jitnmm",
    "jitnmm_p",
]


def _is_static_zero_prob(prob: float, *, op_name: str) -> bool:
    if isinstance(prob, Tracer):
        return False
    prob_arr = np.asarray(prob)
    if prob_arr.size != 1:
        raise ValueError(f"{op_name}: prob must be a scalar, but got shape {prob_arr.shape}.")
    prob_scalar = float(prob_arr.item())
    if not np.isfinite(prob_scalar):
        raise ValueError(f"{op_name}: prob must be finite, but got {prob_scalar}.")
    if not (0. <= prob_scalar <= 1.):
        raise ValueError(f"{op_name}: prob must be in [0, 1], but got {prob_scalar}.")
    return prob_scalar == 0.


def jitn(
    w_loc: Data,
    w_scale: Data,
    prob: float,
    seed: int,
    *,
    shape: MatrixShape,
    transpose: bool = False,
    corder: bool = True,
    backend: Optional[str] = None,
) -> Data:
    """Materialise a JIT normally-distributed random connectivity matrix.

    Generates a dense matrix of shape *shape* (or its transpose) where each
    element is drawn from ``Normal(w_loc, w_scale)`` with independent
    Bernoulli masking at probability *prob*.

    Parameters
    ----------
    w_loc : scalar or Quantity
        Mean of the normal weight distribution.
    w_scale : scalar or Quantity
        Standard deviation of the normal weight distribution.  Must have
        the same physical unit as *w_loc*.
    prob : float
        Connection probability in ``[0, 1]``.
    seed : int
        RNG seed for reproducible connectivity.
    shape : tuple of int
        Logical matrix shape ``(n_pre, n_post)``.
    transpose : bool, optional
        If ``True``, return the transposed matrix of shape
        ``(n_post, n_pre)``.  Default is ``False``.
    corder : bool, optional
        If ``True`` (default), iterate in column-major order internally.
    backend : str or None, optional
        Compute backend (e.g. ``'numba'``, ``'warp'``, ``'pallas'``).

    Returns
    -------
    jax.Array or Quantity
        Dense matrix of shape ``shape`` (or ``shape[::-1]`` when
        *transpose* is ``True``).

    Raises
    ------
    ValueError
        If ``prob`` is not a scalar, is not finite, or is outside ``[0, 1]``.

    See Also
    --------
    jitnmv : Matrix-vector multiply without materialising the matrix.
    jitnmm : Matrix-matrix multiply without materialising the matrix.
    jits : Scalar-weight variant (all non-zeros share one weight).
    jitu : Uniform-weight variant.

    Notes
    -----
    Each entry ``W[i, j]`` of the generated matrix follows the model:

        ``W[i, j] = N(w_loc, w_scale) * B[i, j]``

    where ``N(w_loc, w_scale)`` is a draw from a normal distribution and
    ``B[i, j] ~ Bernoulli(prob)`` is a binary mask.  Equivalently:

    - ``W[i, j] ~ Normal(w_loc, w_scale)`` with probability ``prob``
    - ``W[i, j] = 0`` with probability ``1 - prob``

    The expected value of each entry is ``E[W[i, j]] = prob * w_loc``.

    The connectivity pattern and normal variates are fully determined by
    ``seed`` and ``prob``.  Using the same ``seed`` always produces the
    same matrix.

    This function materialises the full dense matrix.  For implicit
    (non-materialised) products, use :func:`jitnmv` or :func:`jitnmm`.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._jit_normal.float import jitn
        >>> W = jitn(0.0, 1.0, prob=0.1, seed=42, shape=(100, 50))
        >>> W.shape
        (100, 50)
    """
    u.fail_for_dimension_mismatch(w_loc, w_scale, "w_loc and w_scale must have the same dimension.")
    w_loc, unitd = u.split_mantissa_unit(w_loc)
    w_scale = u.Quantity(w_scale).to(unitd).mantissa
    out_dtype = jnp.asarray(w_loc).dtype
    if _is_static_zero_prob(prob, op_name="jitn"):
        out_shape = shape[::-1] if transpose else shape
        return u.maybe_decimal(jnp.zeros(out_shape, dtype=out_dtype) * unitd)
    clen = _initialize_conn_length(prob)
    res = jitn_p_call(
        w_loc,
        w_scale,
        clen,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=backend,
    )[0]
    return u.maybe_decimal(res * unitd)


def jitnmv(
    w_loc: Data,
    w_scale: Data,
    prob: float,
    vector: Data,
    seed: Optional[int] = None,
    *,
    shape: MatrixShape,
    transpose: bool = False,
    corder: bool = True,
    backend: Optional[str] = None,
) -> Data:
    """JIT normally-distributed matrix-vector product.

    Computes ``W @ v`` (or ``W.T @ v``) where ``W`` is a random matrix
    with entries drawn from ``Normal(w_loc, w_scale)`` masked by
    Bernoulli(*prob*), without materialising ``W``.

    Parameters
    ----------
    w_loc : scalar or Quantity
        Mean of the normal weight distribution.
    w_scale : scalar or Quantity
        Standard deviation.  Must share units with *w_loc*.
    prob : float
        Connection probability in ``[0, 1]``.
    vector : jax.Array or Quantity
        Input vector of shape ``(k,)`` where ``k`` equals ``shape[0]``
        when *transpose* is ``True``, or ``shape[1]`` otherwise.
    seed : int or None, optional
        RNG seed.  ``None`` generates a random seed.
    shape : tuple of int
        Logical matrix shape ``(n_pre, n_post)``.
    transpose : bool, optional
        If ``True``, multiply by the transpose.  Default is ``False``.
    corder : bool, optional
        Column-major iteration order.  Default is ``True``.
    backend : str or None, optional
        Compute backend.

    Returns
    -------
    jax.Array or Quantity
        Output vector.  Shape is ``(shape[1],)`` when *transpose* is
        ``True``, or ``(shape[0],)`` otherwise.

    Raises
    ------
    ValueError
        If ``prob`` is not a scalar, is not finite, or is outside ``[0, 1]``.
    ValueError
        If *vector* is not 1-D or its length does not match the matrix shape.

    See Also
    --------
    jitn : Materialise the full matrix.
    jitnmm : Matrix-matrix multiply variant.

    Notes
    -----
    The connectivity matrix ``W`` of shape ``(m, n)`` follows the model:

        ``W[i, j] = N(w_loc, w_scale) * B[i, j]``

    where ``N(w_loc, w_scale)`` is a normal draw and
    ``B[i, j] ~ Bernoulli(prob)`` is a binary mask, both determined by
    ``seed``.

    The matrix-vector product computes:

        ``y[i] = sum_{j=0}^{n-1} W[i, j] * v[j]``

    When ``transpose=True``, the operation becomes ``y = W^T @ v``:

        ``y[j] = sum_{i=0}^{m-1} W[i, j] * v[i]``

    The matrix is never materialised; weights are generated and consumed
    on the fly, avoiding ``O(m * n)`` memory.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._jit_normal.float import jitnmv
        >>> v = jnp.ones(50)
        >>> y = jitnmv(0.0, 1.0, 0.1, v, seed=42, shape=(100, 50))
        >>> y.shape
        (100,)
    """
    u.fail_for_dimension_mismatch(w_loc, w_scale, "w_loc and w_scale must have the same dimension.")
    seed = _initialize_seed(seed)
    w_loc, unitd = u.split_mantissa_unit(w_loc)
    w_scale = u.Quantity(w_scale).to(unitd).mantissa
    vector, unitv = u.split_mantissa_unit(vector)
    out_dtype = jnp.asarray(w_loc).dtype
    if _is_static_zero_prob(prob, op_name="jitnmv"):
        if vector.ndim != 1:
            raise ValueError(f"jitnmv: vector must be 1D, but got {vector.ndim}D.")
        expected = shape[0] if transpose else shape[1]
        if vector.shape[0] != expected:
            raise ValueError(
                f"jitnmv: shape mismatch, got matrix shape {shape} and vector shape {vector.shape}."
            )
        out_size = shape[1] if transpose else shape[0]
        return u.maybe_decimal(jnp.zeros((out_size,), dtype=out_dtype) * unitd * unitv)
    clen = _initialize_conn_length(prob)
    res = jitnmv_p_call(
        w_loc,
        w_scale,
        clen,
        vector,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=backend,
    )[0]
    return u.maybe_decimal(res * unitd * unitv)


def jitnmm(
    w_loc: Data,
    w_scale: Data,
    prob: float,
    B: Data,
    seed: Optional[int] = None,
    *,
    shape: MatrixShape,
    transpose: bool = False,
    corder: bool = True,
    backend: Optional[str] = None,
) -> Data:
    """JIT normally-distributed matrix-matrix product.

    Computes ``W @ B`` (or ``W.T @ B``) where ``W`` is a random matrix
    with entries drawn from ``Normal(w_loc, w_scale)`` masked by
    Bernoulli(*prob*), without materialising ``W``.

    Parameters
    ----------
    w_loc : scalar or Quantity
        Mean of the normal weight distribution.
    w_scale : scalar or Quantity
        Standard deviation.  Must share units with *w_loc*.
    prob : float
        Connection probability in ``[0, 1]``.
    B : jax.Array or Quantity
        Right-hand matrix of shape ``(k, n)`` where ``k`` equals
        ``shape[0]`` when *transpose* is ``True``, or ``shape[1]``
        otherwise.
    seed : int or None, optional
        RNG seed.  ``None`` generates a random seed.
    shape : tuple of int
        Logical matrix shape ``(n_pre, n_post)``.
    transpose : bool, optional
        If ``True``, multiply by the transpose.  Default is ``False``.
    corder : bool, optional
        Column-major iteration order.  Default is ``True``.
    backend : str or None, optional
        Compute backend.

    Returns
    -------
    jax.Array or Quantity
        Output matrix of shape ``(shape[1], n)`` when *transpose* is
        ``True``, or ``(shape[0], n)`` otherwise.

    Raises
    ------
    ValueError
        If ``prob`` is not a scalar, is not finite, or is outside ``[0, 1]``.
    ValueError
        If *B* is not 2-D or its leading dimension does not match the
        matrix shape.

    See Also
    --------
    jitn : Materialise the full matrix.
    jitnmv : Matrix-vector multiply variant.

    Notes
    -----
    The connectivity matrix ``W`` of shape ``(m, n)`` follows the model:

        ``W[i, j] = N(w_loc, w_scale) * B_mask[i, j]``

    where ``N(w_loc, w_scale)`` is a normal draw and
    ``B_mask[i, j] ~ Bernoulli(prob)`` is a binary mask, both determined
    by ``seed``.

    The matrix-matrix product computes:

        ``Y[i, c] = sum_{j=0}^{n-1} W[i, j] * B[j, c]``

    When ``transpose=True``, the operation becomes ``Y = W^T @ B``:

        ``Y[j, c] = sum_{i=0}^{m-1} W[i, j] * B[i, c]``

    The matrix ``W`` is never materialised; weights are generated and
    consumed on the fly.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._jit_normal.float import jitnmm
        >>> B = jnp.ones((50, 10))
        >>> Y = jitnmm(0.0, 1.0, 0.1, B, seed=42, shape=(100, 50))
        >>> Y.shape
        (100, 10)
    """
    u.fail_for_dimension_mismatch(w_loc, w_scale, "w_loc and w_scale must have the same dimension.")
    seed = _initialize_seed(seed)
    w_loc, unitd = u.split_mantissa_unit(w_loc)
    w_scale = u.Quantity(w_scale).to(unitd).mantissa
    B, unitB = u.split_mantissa_unit(B)
    out_dtype = jnp.asarray(w_loc).dtype
    if _is_static_zero_prob(prob, op_name="jitnmm"):
        if B.ndim != 2:
            raise ValueError(f"jitnmm: B must be 2D, but got {B.ndim}D.")
        expected = shape[0] if transpose else shape[1]
        if B.shape[0] != expected:
            raise ValueError(
                f"jitnmm: shape mismatch, got matrix shape {shape} and B shape {B.shape}."
            )
        out_shape = (shape[1], B.shape[1]) if transpose else (shape[0], B.shape[1])
        return u.maybe_decimal(jnp.zeros(out_shape, dtype=out_dtype) * unitd * unitB)
    clen = _initialize_conn_length(prob)
    res = jitnmm_p_call(
        w_loc,
        w_scale,
        clen,
        B,
        seed,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=backend,
    )[0]
    return u.maybe_decimal(res * unitd * unitB)


def _jitn_numba_kernel_generator(corder: bool = True, **kwargs):
    import numba
    _lfsr_seed = get_numba_lfsr_seed()
    _lfsr_random_integers = get_numba_lfsr_random_integers()
    _lfsr_normal = get_numba_lfsr_normal()

    if corder:
        # JIT matrix
        # - JIT matrix shape = [m, n]
        @numba.njit(fastmath=True)
        def kernel(w_loc, w_scale, clen, seed, posts):
            posts[:] = 0.
            m = posts.shape[0]
            n = posts.shape[1]
            w_loc0 = w_loc[0]
            w_scale0 = w_scale[0]
            seed0 = seed[0]
            clen0 = clen[0]
            for i_row in range(m):
                state = _lfsr_seed(seed0 + i_row * n)
                i_col = _lfsr_random_integers(state, 0, clen0 - 1)
                while i_col < n:
                    posts[i_row, i_col] = _lfsr_normal(state, w_loc0, w_scale0)
                    i_col += _lfsr_random_integers(state, 1, clen0 - 1)

    else:
        # JIT matrix
        # - JIT matrix shape = [m, n]
        @numba.njit(fastmath=True)
        def kernel(w_loc, w_scale, clen, seed, posts):
            posts[:] = 0.
            m = posts.shape[0]
            n = posts.shape[1]
            w_loc0 = w_loc[0]
            w_scale0 = w_scale[0]
            clen0 = clen[0]
            seed0 = seed[0]
            for i_col in range(n):
                state = _lfsr_seed(seed0 + i_col * m)
                i_row = _lfsr_random_integers(state, 0, clen0 - 1)
                while i_row < m:
                    posts[i_row, i_col] = _lfsr_normal(state, w_loc0, w_scale0)
                    i_row += _lfsr_random_integers(state, 1, clen0 - 1)

    def run(w_loc, w_scale, clen, seed):
        return numba_kernel(kernel, outs=kwargs['outs'])(w_loc, w_scale, clen, seed)

    return run


def _jitn_warp_kernel_generator(
    w_loc_info: jax.ShapeDtypeStruct,
    w_scale_info: jax.ShapeDtypeStruct,
    clen_info: jax.ShapeDtypeStruct,
    seed_info: jax.ShapeDtypeStruct,
    out_info: jax.ShapeDtypeStruct,
    corder: bool = True,
    **kwargs
):
    import warp
    from warp.jax_experimental import jax_kernel

    w_loc_warp_info = jaxinfo_to_warpinfo(w_loc_info)
    w_scale_warp_info = jaxinfo_to_warpinfo(w_scale_info)
    clen_warp_info = jaxinfo_to_warpinfo(clen_info)
    seed_warp_info = jaxinfo_to_warpinfo(seed_info)
    out_warp_info = jaxinfo_to_warpinfo(out_info)

    if corder:
        # JIT matrix - JIT matrix shape = [m, n]
        @warp.kernel
        def kernel(
            w_loc: w_loc_warp_info,
            w_scale: w_scale_warp_info,
            clen: clen_warp_info,
            seed: seed_warp_info,
            posts: out_warp_info,
        ):
            n = posts.shape[1]
            w_loc0 = w_loc[0]
            w_scale0 = w_scale[0]
            clen0 = clen[0]
            seed0 = seed[0]
            i_row = warp.tid()
            state = warp.rand_init(seed0 + i_row * n)
            i_col = warp.randi(state, 0, clen0)
            while i_col < n:
                posts[i_row, i_col] = warp.randn(state) * w_scale0 + w_loc0
                i_col += warp.randi(state, 1, clen0)


    else:
        # JIT matrix - JIT matrix shape = [m, n]
        @warp.kernel
        def kernel(
            w_loc: w_loc_warp_info,
            w_scale: w_scale_warp_info,
            clen: clen_warp_info,
            seed: seed_warp_info,
            posts: out_warp_info,
        ):
            m = posts.shape[0]
            w_loc0 = w_loc[0]
            w_scale0 = w_scale[0]
            clen0 = clen[0]
            seed0 = seed[0]
            i_col = warp.tid()
            state = warp.rand_init(seed0 + i_col * m)
            i_row = warp.randi(state, 0, clen0)
            while i_row < m:
                posts[i_row, i_col] = warp.randn(state) * w_scale0 + w_loc0
                i_row += warp.randi(state, 1, clen0)

    def run(w_loc, w_scale, clen, seed):
        dim = out_info.shape[0] if corder else out_info.shape[1]
        fn = jax_kernel(kernel, launch_dims=[dim], num_outputs=1, in_out_argnames=['posts'])
        return fn(w_loc, w_scale, clen, seed, jnp.zeros(out_info.shape, out_info.dtype))

    return run


def _jitn_pallas_kernel_generator(
    out_info: jax.ShapeDtypeStruct,
    corder: bool = True,
    **kwargs
):
    from jax.experimental import pallas as pl
    _PallasLFSRRNG = get_pallas_lfsr_rng_class()

    dim = out_info.shape[0] if corder else out_info.shape[1]
    block_size = generate_block_dim(dim, maximum=128)
    n_block = pl.cdiv(dim, block_size)

    if corder:
        def kernel(w_loc_ref, w_scale_ref, clen_ref, seed_ref, _, post_ref):
            m = post_ref.shape[1]
            w_loc = w_loc_ref[0]
            w_scale = w_scale_ref[0]
            clen0 = clen_ref[0]
            seed0 = seed_ref[0]
            i_row_block = pl.program_id(0)
            i_rows = i_row_block * block_size + jnp.arange(block_size)
            i_row_mask = i_rows < dim

            def body(data):
                i_cols, i_col_mask, rng = data
                val = rng.normal(w_loc, w_scale)
                safe_cols = jnp.where(i_col_mask, i_cols, 0)
                post_ref[i_rows, safe_cols] = jnp.where(i_row_mask & i_col_mask, val, post_ref[i_rows, safe_cols])
                i_cols += rng.random_integers(1, clen0)
                return i_cols, i_cols < m, rng

            rng = _PallasLFSRRNG(seed0 + i_rows * m)
            i_cols = rng.random_integers(0, clen0)
            i_col_mask = i_cols < m
            jax.lax.while_loop(
                lambda data: jnp.sum(data[1]) > 0,
                body,
                (i_cols, i_col_mask, rng)
            )

    else:
        def kernel(w_loc_ref, w_scale_ref, clen_ref, seed_ref, _, post_ref):
            n = post_ref.shape[0]
            w_loc = w_loc_ref[0]
            w_scale = w_scale_ref[0]
            clen0 = clen_ref[0]
            seed0 = seed_ref[0]
            i_col_block = pl.program_id(0)
            i_cols = i_col_block * block_size + jnp.arange(block_size)
            i_col_mask = i_cols < dim

            def body(data):
                i_rows, i_row_mask, rng = data
                val = rng.normal(w_loc, w_scale)
                safe_rows = jnp.where(i_row_mask, i_rows, 0)
                post_ref[safe_rows, i_cols] = jnp.where(i_row_mask & i_col_mask, val, post_ref[safe_rows, i_cols])
                i_rows = i_rows + rng.random_integers(1, clen0)
                return i_rows, i_rows < n, rng

            rng = _PallasLFSRRNG(seed0 + i_cols * n)
            i_rows = rng.random_integers(0, clen0)
            i_row_mask = i_rows < n
            jax.lax.while_loop(
                lambda data: jnp.sum(data[1]) > 0,
                body,
                (i_rows, i_row_mask, rng)
            )

    def run(w_loc, w_scale, clen, seed):
        fn = pl.pallas_call(kernel, grid=(n_block,), input_output_aliases={4: 0}, out_shape=kwargs['outs'],
                            backend='triton')
        out = kwargs['outs'][0]
        placeholder = jnp.zeros(out.shape, out.dtype)
        return fn(w_loc, w_scale, clen, seed, placeholder)

    return run


def _jitn_jvp_wlow(w_loc_dot, w_loc, w_scale, clen, seed, *, out_info, **kwargs):
    out = jnp.ones_like(out_info) * w_loc_dot
    return [out]


def _jitn_jvp_whigh(
    w_scale_dot, w_loc, w_scale, clen, seed, *,
    shape, transpose: bool, corder: bool, backend, **kwargs
):
    return jitn_p_call(
        0., w_scale_dot, clen, seed, shape=shape, transpose=transpose, corder=corder, backend=backend
    )


def _jitn_transpose(
    ct, w_loc, w_scale, clen, seed, *,
    shape, transpose: bool, corder: bool, backend, **kwargs
):
    assert not ad.is_undefined_primal(clen)
    assert not ad.is_undefined_primal(seed)
    ct = ct[0]
    if ad.is_undefined_primal(w_loc):
        dwlow = jnp.expand_dims(ct.sum(), axis=0)
        return (dwlow, w_scale, clen, seed)

    elif ad.is_undefined_primal(w_scale):
        forward = jitn_p_call(
            0., 1., clen, seed, shape=shape, transpose=transpose, corder=corder, backend=backend
        )[0]
        dwscale = jnp.expand_dims((ct * forward).sum(), axis=0)
        return (w_loc, dwscale, clen, seed)

    else:
        raise NotImplementedError(
            'JITC matrix transpose is only implemented for the w_low and w_scale arguments.'
        )


def _jitn_batching(args, axes, **kwargs):
    return general_batching_rule(jitn_p, args, axes, **kwargs)


def _jitn_benchmark_data(*, platform):
    n_pre, n_post, prob, dtype = 1000, 1000, 0.1, jnp.float32
    configs = []
    for transpose in (False, True):
        for corder in (True, False):
            w_loc = jnp.ones(1, dtype=dtype)
            w_scale = jnp.ones(1, dtype=dtype) * 0.1
            clen = jnp.atleast_1d(jnp.asarray(2.0 / prob, dtype=dtype))
            seed = jnp.asarray(42, dtype=jnp.uint32)
            name = f"{'T' if transpose else 'NT'},{'corder' if corder else 'rorder'}"
            configs.append(BenchmarkConfig(name, (w_loc, w_scale, clen, seed), {
                'shape': (n_pre, n_post), 'transpose': transpose, 'corder': corder
            }))
    return configs


def jitn_p_call(
    w_loc, w_scale, clen, seed, *, shape, transpose: bool, corder: bool, backend
):
    """Dispatch the JIT normal matrix materialisation primitive.

    Parameters
    ----------
    w_loc : jax.Array
        Weight mean, shape ``(1,)``.
    w_scale : jax.Array
        Weight standard deviation, shape ``(1,)``.
    clen : jax.Array
        Connection length derived from *prob*, shape ``(1,)``.
    seed : jax.Array
        RNG seed, shape ``(1,)``.
    shape : tuple of int
        Logical matrix shape ``(n_pre, n_post)``.
    transpose : bool
        Whether to transpose the output.
    corder : bool
        Column-major iteration order.
    backend : str or None, optional
        Compute backend.

    Returns
    -------
    tuple of jax.Array
        Single-element tuple containing the materialised matrix.

    See Also
    --------
    jitn : High-level wrapper with unit support.

    Notes
    -----
    Scalar inputs (``w_loc``, ``w_scale``, ``clen``, ``seed``) are
    automatically promoted to 1-D arrays of shape ``(1,)``.

    The generated matrix has entry model:

        ``W[i, j] = Normal(w_loc[0], w_scale[0]) * Bernoulli(prob)``

    where ``prob`` is implicitly encoded through ``clen = 2 / prob``.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._jit_normal.float import jitn_p_call
        >>> out = jitn_p_call(
        ...     jnp.array([0.0]), jnp.array([1.0]),
        ...     jnp.array([20.0]), jnp.array([42]),
        ...     shape=(10, 5), transpose=False, corder=True)
    """
    seed = _initialize_seed(seed)
    w_loc = jnp.atleast_1d(w_loc)
    w_scale = jnp.atleast_1d(w_scale)
    clen = jnp.atleast_1d(clen)
    seed = jnp.atleast_1d(seed)

    out_info = (
        jax.ShapeDtypeStruct(shape[::-1], dtype=w_loc.dtype)
        if transpose else
        jax.ShapeDtypeStruct(shape, dtype=w_loc.dtype)
    )

    return jitn_p(
        w_loc,
        w_scale,
        clen,
        seed,
        outs=[out_info],
        w_loc_info=jax.ShapeDtypeStruct(w_loc.shape, w_loc.dtype),
        w_scale_info=jax.ShapeDtypeStruct(w_scale.shape, w_scale.dtype),
        clen_info=jax.ShapeDtypeStruct(clen.shape, clen.dtype),
        seed_info=jax.ShapeDtypeStruct(seed.shape, seed.dtype),
        out_info=out_info,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=backend,
    )


jitn_p = XLACustomKernel(
    'float_jitn',
    doc="""
Low-level XLA custom-kernel primitive for ``jitn``.

This ``XLACustomKernel`` instance dispatches the JIT normal connectivity matrix generation
operation to registered backends (``numba``, ``warp``, ``pallas``),
using runtime shape/dtype metadata provided by the high-level wrapper.

This operation generates a sparse connectivity matrix where weights are normally distributed
with specified mean and standard deviation. The connectivity pattern is generated on-the-fly
using a deterministic PRNG seeded by the provided seed value.

Beyond backend dispatch, the primitive stores JAX transformation bindings
(JVP, transpose, batching, and call registration) so the operation integrates
correctly with ``jit``, ``vmap``, and autodiff.

Available backends can be queried with ``jitn_p.available_backends(platform)``,
and the default backend can be configured with ``jitn_p.set_default(platform, backend)``.

See Also
--------
jitn : High-level user-facing function wrapper.
"""
)
jitn_p.def_numba_kernel(_jitn_numba_kernel_generator)
jitn_p.def_warp_kernel(_jitn_warp_kernel_generator)
jitn_p.def_pallas_kernel('gpu', _jitn_pallas_kernel_generator)
jitn_p.def_jvp_rule2(_jitn_jvp_wlow, _jitn_jvp_whigh, None, None)
jitn_p.def_transpose_rule(_jitn_transpose)
jitn_p.def_batching_rule(_jitn_batching)
jitn_p.def_tags('jit_normal', 'float')
jitn_p.def_benchmark_data(_jitn_benchmark_data)


def _jitnmv_numba_kernel_generator(
    transpose: bool = False,
    corder: bool = True,
    **kwargs
):
    import numba
    _lfsr_seed = get_numba_lfsr_seed()
    _lfsr_random_integers = get_numba_lfsr_random_integers()
    _lfsr_normal = get_numba_lfsr_normal()

    if corder:
        @numba.njit(fastmath=True)
        def kernel(w_loc, w_scale, clen, vector, seed, posts):
            posts[:] = 0.
            num_row = posts.shape[0]
            num_col = vector.shape[0]
            w_loc0 = w_loc[0]
            w_scale0 = w_scale[0]
            seed0 = seed[0]
            clen0 = clen[0]
            for i_row in range(num_row):
                state = _lfsr_seed(seed0 + i_row * num_col)
                i_col = _lfsr_random_integers(state, 0, clen0 - 1)
                out = np.asarray(0., dtype=vector.dtype)
                while i_col < num_col:
                    out += vector[i_col] * _lfsr_normal(state, w_loc0, w_scale0)
                    i_col += _lfsr_random_integers(state, 1, clen0 - 1)
                posts[i_row] = out

    else:
        @numba.njit(fastmath=True)
        def kernel(w_loc, w_scale, clen, vector, seed, posts):
            posts[:] = 0.
            num_row = posts.shape[0]
            num_col = vector.shape[0]
            w_loc0 = w_loc[0]
            w_scale0 = w_scale[0]
            clen0 = clen[0]
            seed0 = seed[0]
            for i_col in range(num_col):
                state = _lfsr_seed(seed0 + i_col * num_row)
                v = vector[i_col]
                i_row = _lfsr_random_integers(state, 0, clen0 - 1)
                while i_row < num_row:
                    posts[i_row] += v * _lfsr_normal(state, w_loc0, w_scale0)
                    i_row += _lfsr_random_integers(state, 1, clen0 - 1)

    def run(w_loc, w_scale, clen, vector, seed):
        return numba_kernel(kernel, outs=kwargs['outs'])(w_loc, w_scale, clen, vector, seed)

    return run


def _jitnmv_warp_kernel_generator(
    w_loc_info: jax.ShapeDtypeStruct,
    w_scale_info: jax.ShapeDtypeStruct,
    clen_info: jax.ShapeDtypeStruct,
    vector_info: jax.ShapeDtypeStruct,
    seed_info: jax.ShapeDtypeStruct,
    out_info: jax.ShapeDtypeStruct,
    corder: bool = True,
    **kwargs
):
    r"""
    Generate the GPU kernel for the :func:`_jitc_matvec_normal` operation.
    """
    import warp
    from warp.jax_experimental import jax_kernel

    w_loc_warp_info = jaxinfo_to_warpinfo(w_loc_info)
    w_scale_warp_info = jaxinfo_to_warpinfo(w_scale_info)
    clen_warp_info = jaxinfo_to_warpinfo(clen_info)
    v_warp_info = jaxinfo_to_warpinfo(vector_info)
    seed_warp_info = jaxinfo_to_warpinfo(seed_info)
    out_warp_info = jaxinfo_to_warpinfo(out_info)

    if corder:
        @warp.kernel
        def kernel(
            w_loc: w_loc_warp_info,
            w_scale: w_scale_warp_info,
            clen: clen_warp_info,
            vector: v_warp_info,
            seed: seed_warp_info,
            posts: out_warp_info,
        ):
            num_col = vector.shape[0]
            w_loc0 = w_loc[0]
            w_scale0 = w_scale[0]
            clen0 = clen[0]
            seed0 = seed[0]
            i_row = warp.tid()
            r = float(0.0)
            state = warp.rand_init(seed0 + i_row * num_col)
            i_col = warp.randi(state, 0, clen0)
            while i_col < num_col:
                w = warp.randn(state) * w_scale0 + w_loc0
                r += vector[i_col] * w
                i_col += warp.randi(state, 1, clen0)
            posts[i_row] = r

    else:
        @warp.kernel
        def kernel(
            w_loc: w_loc_warp_info,
            w_scale: w_scale_warp_info,
            clen: clen_warp_info,
            vector: v_warp_info,
            seed: seed_warp_info,
            posts: out_warp_info,
        ):
            num_row = posts.shape[0]
            w_loc0 = w_loc[0]
            w_scale0 = w_scale[0]
            clen0 = clen[0]
            seed0 = seed[0]
            i_col = warp.tid()
            v = vector[i_col]
            state = warp.rand_init(seed0 + i_col * num_row)
            i_row = warp.randi(state, 0, clen0)
            while i_row < num_row:
                w = warp.randn(state) * w_scale0 + w_loc0
                warp.atomic_add(posts, i_row, v * w)
                i_row += warp.randi(state, 1, clen0)

    def run(w_loc, w_scale, clen, vector, seed):
        dim = out_info.shape[0] if corder else vector_info.shape[0]
        fn = jax_kernel(kernel, launch_dims=[dim], num_outputs=1, in_out_argnames=['posts'])
        return fn(w_loc, w_scale, clen, vector, seed, jnp.zeros(out_info.shape, out_info.dtype))

    return run


def _jitnmv_pallas_kernel_generator(
    vector_info: jax.ShapeDtypeStruct,
    out_info: jax.ShapeDtypeStruct,
    corder: bool = True,
    **kwargs
):
    from jax.experimental import pallas as pl
    from jax.experimental.pallas.triton import atomic_add  # type: ignore[assignment]
    _PallasLFSRRNG = get_pallas_lfsr_rng_class()

    dim = (out_info.shape[0] if corder else vector_info.shape[0])
    block_size = generate_block_dim(dim, maximum=128)

    if corder:
        def kernel(w_loc_ref, w_scale_ref, clen_ref, vector_ref, seed_ref, _, post_ref):
            num_row = vector_ref.shape[0]
            w_loc = w_loc_ref[0]
            w_scale = w_scale_ref[0]
            clen = clen_ref[0]
            seed = seed_ref[0]
            i_col_block = pl.program_id(0)
            i_cols = i_col_block * block_size + jnp.arange(block_size)
            i_col_mask = i_cols < dim

            def body(data):
                i_rows, i_row_mask, rng, out = data
                v = jnp.where(i_row_mask, vector_ref[i_rows], 0.)
                out += v * rng.normal(w_loc, w_scale)
                i_rows += rng.random_integers(1, clen)
                return i_rows, i_rows < num_row, rng, out

            rng = _PallasLFSRRNG(seed + i_cols * num_row)
            i_rows = rng.random_integers(0, clen)
            i_row_mask = i_rows < num_row
            out = jnp.zeros(block_size, dtype=post_ref.dtype)
            out = jax.lax.while_loop(
                lambda data: jnp.sum(data[1]) > 0,
                body,
                (i_rows, i_row_mask, rng, out)
            )[-1]
            post_ref[i_cols] = jnp.where(i_col_mask, out, post_ref[i_cols])

    else:
        def kernel(w_loc_ref, w_scale_ref, clen_ref, vector_ref, seed_ref, _, post_ref):
            num_col = post_ref.shape[0]
            w_loc = w_loc_ref[0]
            w_scale = w_scale_ref[0]
            clen = clen_ref[0]
            seed = seed_ref[0]
            i_row_block = pl.program_id(0)
            i_rows = i_row_block * block_size + jnp.arange(block_size)
            i_row_mask = i_rows < dim
            vector = jnp.where(i_row_mask, vector_ref[i_rows], 0.)

            def body(data):
                i_cols, i_col_mask, rng = data
                atomic_add(post_ref, (i_cols,), vector * rng.normal(w_loc, w_scale), mask=i_row_mask & i_col_mask)
                i_cols += rng.random_integers(1, clen)
                return i_cols, i_cols < num_col, rng

            rng = _PallasLFSRRNG(seed + i_rows * num_col)
            i_cols = rng.random_integers(0, clen)
            i_col_mask = i_cols < num_col
            jax.lax.while_loop(
                lambda data: jnp.sum(data[1]) > 0,
                body,
                (i_cols, i_col_mask, rng)
            )

    def run(w_loc, w_scale, clen, vector, seed):
        fn = pl.pallas_call(
            kernel,
            grid=(pl.cdiv(dim, block_size),),
            input_output_aliases={5: 0},
            out_shape=kwargs['outs'],
            backend='triton',
        )
        placeholder = jnp.zeros(kwargs['outs'][0].shape, kwargs['outs'][0].dtype)
        return fn(w_loc, w_scale, clen, vector, seed, placeholder)

    return run


def _jitnmv_jvp_v(v_dot, w_loc, w_scale, clen, vector, seed, *, shape, transpose, corder, backend, **kwargs):
    return jitnmv_p_call(
        w_loc, w_scale, clen, v_dot, seed, shape=shape, transpose=transpose, corder=corder, backend=backend
    )


def _jitnmv_jvp_wloc(w_dot, w_loc, w_scale, clen, vector, seed, *, shape, transpose, corder, backend, **kwargs):
    return jitnmv_p_call(
        w_dot, w_scale, clen, vector, seed, shape=shape, transpose=transpose, corder=corder, backend=backend
    )


def _jitnmv_jvp_wscale(w_dot, w_loc, w_scale, clen, vector, seed, *, shape, transpose, corder, backend, **kwargs):
    return jitnmv_p_call(
        w_loc, w_dot, clen, vector, seed, shape=shape, transpose=transpose, corder=corder, backend=backend
    )


def _jitnmv_transpose_rules(
    ct, w_loc, w_scale, clen, vector, seed, *, shape, transpose, corder, backend, **kwargs
):
    assert not ad.is_undefined_primal(clen)
    assert not ad.is_undefined_primal(seed)

    ct = ct[0]
    if ad.is_undefined_primal(vector):
        # Gradient w.r.t. vector: d(loss)/d(v) = M^T @ ct
        r = jitnmv_p_call(
            w_loc,
            w_scale,
            clen,
            ct,
            seed,
            shape=shape,
            transpose=not transpose,
            corder=not corder,
            backend=backend,
        )[0]
        return w_loc, w_scale, clen, r, seed
    elif ad.is_undefined_primal(w_loc):
        # M = (w_loc + w_scale * Z) * mask
        # d(M @ v)/d(w_loc) = mask @ v
        # d(loss)/d(w_loc) = ct^T @ (mask @ v) = (mask^T @ ct) . v
        # mask^T @ ct = jitnmv(1., 0., ...) with transposed shape
        r = jitnmv_p_call(
            1., 0., clen, ct, seed,
            shape=shape, transpose=not transpose, corder=not corder,
            backend=backend,
        )[0]
        dw_loc = jnp.expand_dims(jnp.sum(r * vector), axis=0)
        return dw_loc, w_scale, clen, vector, seed
    elif ad.is_undefined_primal(w_scale):
        # M = (w_loc + w_scale * Z) * mask
        # d(M @ v)/d(w_scale) = (Z * mask) @ v
        # d(loss)/d(w_scale) = ct^T @ ((Z * mask) @ v) = ((Z * mask)^T @ ct) . v
        # (Z * mask)^T @ ct = jitnmv(0., 1., ...) with transposed shape
        r = jitnmv_p_call(
            0., 1., clen, ct, seed,
            shape=shape, transpose=not transpose, corder=not corder,
            backend=backend,
        )[0]
        dw_scale = jnp.expand_dims(jnp.sum(r * vector), axis=0)
        return w_loc, dw_scale, clen, vector, seed
    else:
        raise NotImplementedError(
            f"Transpose rule for jitnmv not implemented "
            f"when none of vector/w_loc/w_scale is an undefined primal."
        )


def _jitnmv_batching(args, axes, **kwargs):
    if tuple(axes) == (None, None, None, 0, None):
        assert args[3].ndim == 2, 'Batching axis 0 requires 2D input.'
        r = jitnmm_p_call(
            args[0],
            args[1],
            args[2],
            args[3].T,
            args[4],
            shape=kwargs['shape'],
            transpose=kwargs['transpose'],
            corder=kwargs['corder'],
            backend=kwargs.get('backend'),
        )
        return r, [1]
    elif tuple(axes) == (None, None, None, 1, None):
        assert args[3].ndim == 2, 'Batching axis 0 requires 2D input.'
        r = jitnmm_p_call(
            args[0],
            args[1],
            args[2],
            args[3],
            args[4],
            shape=kwargs['shape'],
            transpose=kwargs['transpose'],
            corder=kwargs['corder'],
            backend=kwargs.get('backend'),
        )
        return r, [1]
    else:
        return general_batching_rule(
            jitnmv_p,
            args,
            axes,
            **kwargs,
        )


def _jitnmv_benchmark_data(*, platform):
    n_pre, n_post, prob, dtype = 1000, 1000, 0.1, jnp.float32
    configs = []
    for transpose in (False, True):
        for corder in (True, False):
            w_loc = jnp.ones(1, dtype=dtype)
            w_scale = jnp.ones(1, dtype=dtype) * 0.1
            clen = jnp.atleast_1d(jnp.asarray(2.0 / prob, dtype=dtype))
            v_size = n_post if not transpose else n_pre
            vector = jnp.asarray(np.random.randn(v_size), dtype=dtype)
            seed = jnp.asarray(42, dtype=jnp.uint32)
            name = f"{'T' if transpose else 'NT'},{'corder' if corder else 'rorder'}"
            configs.append(
                BenchmarkConfig(
                    name,
                    (w_loc, w_scale, clen, vector, seed),
                    {'shape': (n_pre, n_post), 'transpose': transpose, 'corder': corder}
                )
            )
    return configs


def jitnmv_p_call(
    w_loc,
    w_scale,
    clen,
    vector,
    seed,
    *,
    shape,
    transpose: bool,
    corder: bool,
    backend,
):
    """Dispatch the JIT normal matrix-vector multiply primitive.

    Parameters
    ----------
    w_loc : jax.Array
        Weight mean, shape ``(1,)``.
    w_scale : jax.Array
        Weight standard deviation, shape ``(1,)``.
    clen : jax.Array
        Connection length, shape ``(1,)``.
    vector : jax.Array
        Input 1-D vector.
    seed : jax.Array
        RNG seed, shape ``(1,)``.
    shape : tuple of int
        Logical matrix shape ``(n_pre, n_post)``.
    transpose : bool
        Whether to use the transposed matrix.
    corder : bool
        Column-major iteration order.
    backend : str or None, optional
        Compute backend.

    Returns
    -------
    tuple of jax.Array
        Single-element tuple containing the output vector.

    Raises
    ------
    AssertionError
        If ``shape`` is not length 2, or if ``w_loc``, ``w_scale``,
        ``clen``, ``seed`` do not have shape ``(1,)``, or if ``vector``
        is not 1-D, or if the matrix shape and vector length are
        incompatible.

    See Also
    --------
    jitnmv : High-level wrapper with unit support.

    Notes
    -----
    The product is computed without materialising the full matrix:

        ``y[i] = sum_{j} Normal(w_loc, w_scale) * Bernoulli(prob) * v[j]``

    Each weight is generated on the fly using a deterministic PRNG seeded
    by ``seed``.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._jit_normal.float import jitnmv_p_call
        >>> out = jitnmv_p_call(
        ...     jnp.array([0.0]), jnp.array([1.0]),
        ...     jnp.array([20.0]), jnp.ones(5), jnp.array([42]),
        ...     shape=(10, 5), transpose=False, corder=True)
    """
    w_loc = jnp.atleast_1d(w_loc)
    w_scale = jnp.atleast_1d(w_scale)
    clen = jnp.atleast_1d(clen)

    assert len(shape) == 2, "The matrix shape should be a tuple of two integers."
    assert w_loc.shape == (1,), f"The weight shape should be (1,), but got {w_loc.shape}."
    assert w_scale.shape == (1,), f"The weight shape should be (1,), but got {w_scale.shape}."
    assert clen.shape == (1,), f"The clen shape should be (1,), but got {clen.shape}."
    assert vector.ndim == 1, f"The vector should be a 1D array, but got {vector.ndim}D."
    assert seed.shape == (1,), f"The seed shape should be (1,), but got {seed.shape}."

    if transpose:
        assert shape[0] == len(vector), f"The matrix shape and vector length do not match. {vector.shape} @ {shape}"
    else:
        assert shape[1] == len(vector), f"The matrix shape and vector length do not match. {shape} @ {vector.shape}"

    out_info = (
        jax.ShapeDtypeStruct([shape[1]], w_loc.dtype)
        if transpose else
        jax.ShapeDtypeStruct([shape[0]], w_loc.dtype)
    )

    return jitnmv_p(
        w_loc,
        w_scale,
        clen,
        vector,
        seed,
        outs=[out_info],
        w_loc_info=jax.ShapeDtypeStruct(w_loc.shape, w_loc.dtype),
        w_scale_info=jax.ShapeDtypeStruct(w_scale.shape, w_scale.dtype),
        clen_info=jax.ShapeDtypeStruct(clen.shape, clen.dtype),
        vector_info=jax.ShapeDtypeStruct(vector.shape, vector.dtype),
        seed_info=jax.ShapeDtypeStruct(seed.shape, seed.dtype),
        out_info=out_info,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=backend,
    )


jitnmv_p = XLACustomKernel(
    'float_jitnmv',
    doc="""
Low-level XLA custom-kernel primitive for ``jitnmv``.

This ``XLACustomKernel`` instance dispatches the JIT normal connectivity matrix-vector
multiplication with floating-point weights operation to registered backends
(``numba``, ``warp``, ``pallas``), using runtime shape/dtype metadata provided by
the high-level wrapper.

In this operation, the connectivity matrix has weights normally distributed with specified
mean and standard deviation, and the input vector contains floating-point values. The
operation computes a standard matrix-vector product without event-driven sparsity.

Beyond backend dispatch, the primitive stores JAX transformation bindings
(JVP, transpose, batching, and call registration) so the operation integrates
correctly with ``jit``, ``vmap``, and autodiff.

Available backends can be queried with ``jitnmv_p.available_backends(platform)``,
and the default backend can be configured with ``jitnmv_p.set_default(platform, backend)``.

See Also
--------
jitnmv : High-level user-facing function wrapper.
"""
)
jitnmv_p.def_numba_kernel(_jitnmv_numba_kernel_generator)
jitnmv_p.def_warp_kernel(_jitnmv_warp_kernel_generator)
jitnmv_p.def_pallas_kernel('gpu', _jitnmv_pallas_kernel_generator)
jitnmv_p.def_jvp_rule2(_jitnmv_jvp_wloc, _jitnmv_jvp_wscale, None, _jitnmv_jvp_v, None)
jitnmv_p.def_transpose_rule(_jitnmv_transpose_rules)
jitnmv_p.def_batching_rule(_jitnmv_batching)
jitnmv_p.def_tags('jit_normal', 'float')
jitnmv_p.def_benchmark_data(_jitnmv_benchmark_data)


def _jitnmm_numba_kernel_generator(
    corder: bool = True,
    **kwargs
):
    import numba
    _lfsr_seed = get_numba_lfsr_seed()
    _lfsr_random_integers = get_numba_lfsr_random_integers()
    _lfsr_normal = get_numba_lfsr_normal()

    if corder:
        # JIT Matrix @ B
        # - JIT matrix: [m, k]
        # - B: [k, n]
        @numba.njit(fastmath=True)
        def kernel(w_loc, w_scale, clen, B, seed, posts):
            posts[:] = 0.
            m = posts.shape[0]
            n = posts.shape[1]
            k = B.shape[0]
            w_loc0 = w_loc[0]
            w_scale0 = w_scale[0]
            seed0 = seed[0]
            clen0 = clen[0]
            for i_m in range(m):
                state = _lfsr_seed(seed0 + i_m * k)
                i_k = _lfsr_random_integers(state, 0, clen0 - 1)
                out = np.zeros(n, dtype=B.dtype)
                while i_k < k:
                    out += B[i_k] * _lfsr_normal(state, w_loc0, w_scale0)
                    i_k += _lfsr_random_integers(state, 1, clen0 - 1)
                posts[i_m] = out

    else:
        # JIT Matrix @ B
        # - JIT matrix: [m, k]
        # - B: [k, n]
        @numba.njit(fastmath=True)
        def kernel(w_loc, w_scale, clen, B, seed, posts):
            posts[:] = 0.
            m = posts.shape[0]
            k = B.shape[0]
            w_loc0 = w_loc[0]
            w_scale0 = w_scale[0]
            seed0 = seed[0]
            clen0 = clen[0]
            for i_k in range(k):
                state = _lfsr_seed(seed0 + i_k * m)
                out = B[i_k]
                i_m = _lfsr_random_integers(state, 0, clen0 - 1)
                while i_m < m:
                    posts[i_m] += out * _lfsr_normal(state, w_loc0, w_scale0)
                    i_m += _lfsr_random_integers(state, 1, clen0 - 1)

    def run(w_loc, w_scale, clen, B, seed):
        return numba_kernel(kernel, outs=kwargs['outs'])(w_loc, w_scale, clen, B, seed)

    return run


def _jitnmm_warp_kernel_generator(
    w_loc_info: jax.ShapeDtypeStruct,
    w_scale_info: jax.ShapeDtypeStruct,
    clen_info: jax.ShapeDtypeStruct,
    B_info: jax.ShapeDtypeStruct,
    out_info: jax.ShapeDtypeStruct,
    seed_info: jax.ShapeDtypeStruct,
    corder: bool,
    **kwargs
):
    import warp
    from warp.jax_experimental import jax_kernel

    w_loc_warp_info = jaxinfo_to_warpinfo(w_loc_info)
    w_scale_warp_info = jaxinfo_to_warpinfo(w_scale_info)
    clen_warp_info = jaxinfo_to_warpinfo(clen_info)
    B_warp_info = jaxinfo_to_warpinfo(B_info)
    seed_warp_info = jaxinfo_to_warpinfo(seed_info)
    out_warp_info = jaxinfo_to_warpinfo(out_info)

    # TILE_SIZE = generate_block_dim(B_info.shape[1], maximum=512)
    #
    # if corder:
    #     # JIT Matrix @ B, corder=True
    #     # Each thread i_m generates one row of the JITC matrix and
    #     # multiplies it with B, accumulating into posts[i_m, :].
    #     @warp.kernel
    #     def kernel(
    #         w_loc: w_loc_warp_info,
    #         w_scale: w_scale_warp_info,
    #         clen: clen_warp_info,
    #         B: B_warp_info,
    #         seed: seed_warp_info,
    #         posts: out_warp_info,
    #     ):
    #         k = B.shape[0]
    #         n = B.shape[1]
    #         w_loc0 = w_loc[0]
    #         w_scale0 = w_scale[0]
    #         clen0 = clen[0]
    #         seed0 = seed[0]
    #
    #         i_m = warp.tid()
    #         state = warp.rand_init(seed0 + i_m * k)
    #         i_k = warp.randi(state, 0, clen0)
    #         while i_k < k:
    #             w = warp.randn(state) * w_scale0 + w_loc0
    #             for j in range(n):
    #                 posts[i_m, j] += B[i_k, j] * w
    #             i_k += warp.randi(state, 1, clen0)
    #
    # else:
    #     # JIT Matrix @ B, corder=False
    #     # Each thread i_k generates one column of the JITC matrix and
    #     # scatters B[i_k, :] scaled by weight into output rows via atomic adds.
    #     @warp.kernel
    #     def kernel(
    #         w_loc: w_loc_warp_info,
    #         w_scale: w_scale_warp_info,
    #         clen: clen_warp_info,
    #         B: B_warp_info,
    #         seed: seed_warp_info,
    #         posts: out_warp_info,
    #     ):
    #         m = posts.shape[0]
    #         n = B.shape[1]
    #         w_loc0 = w_loc[0]
    #         w_scale0 = w_scale[0]
    #         clen0 = clen[0]
    #         seed0 = seed[0]
    #
    #         i_k = warp.tid()
    #         state = warp.rand_init(seed0 + i_k * m)
    #         i_m = warp.randi(state, 0, clen0)
    #         while i_m < m:
    #             w = warp.randn(state) * w_scale0 + w_loc0
    #             for j in range(n):
    #                 warp.atomic_add(posts, i_m, j, B[i_k, j] * w)
    #             i_m += warp.randi(state, 1, clen0)
    #
    # def run(w_loc, w_scale, clen, B, seed):
    #     dim = out_info.shape[0] if corder else B_info.shape[0]
    #     fn = jax_kernel(kernel, launch_dims=[dim], num_outputs=1, in_out_argnames=['posts'])
    #     return fn(w_loc, w_scale, clen, B, seed, jnp.zeros(out_info.shape, out_info.dtype))

    if corder:
        # JIT Matrix @ B, corder=True
        # Each thread i_m generates one row of the JITC matrix and
        # multiplies it with B, accumulating into posts[i_m, :].
        @warp.kernel
        def kernel(
            w_loc: w_loc_warp_info,
            w_scale: w_scale_warp_info,
            clen: clen_warp_info,
            B: B_warp_info,
            seed: seed_warp_info,
            posts: out_warp_info,
        ):
            k = B.shape[0]
            n = B.shape[1]
            w_loc0 = w_loc[0]
            w_scale0 = w_scale[0]
            clen0 = clen[0]
            seed0 = seed[0]

            i_m = warp.tid()
            state = warp.rand_init(seed0 + i_m * k)
            i_k = warp.randi(state, 0, clen0)
            while i_k < k:
                w = warp.randn(state) * w_scale0 + w_loc0
                for j in range(n):
                    posts[i_m, j] += B[i_k, j] * w
                i_k += warp.randi(state, 1, clen0)

    else:
        # JIT Matrix @ B, corder=False
        # Each thread i_k generates one column of the JITC matrix and
        # scatters B[i_k, :] scaled by weight into output rows via atomic adds.
        @warp.kernel
        def kernel(
            w_loc: w_loc_warp_info,
            w_scale: w_scale_warp_info,
            clen: clen_warp_info,
            B: B_warp_info,
            seed: seed_warp_info,
            posts: out_warp_info,
        ):
            m = posts.shape[0]
            n = B.shape[1]
            w_loc0 = w_loc[0]
            w_scale0 = w_scale[0]
            clen0 = clen[0]
            seed0 = seed[0]

            i_k = warp.tid()
            state = warp.rand_init(seed0 + i_k * m)
            i_m = warp.randi(state, 0, clen0)
            while i_m < m:
                w = warp.randn(state) * w_scale0 + w_loc0
                for j in range(n):
                    warp.atomic_add(posts, i_m, j, B[i_k, j] * w)
                i_m += warp.randi(state, 1, clen0)

    def run(w_loc, w_scale, clen, B, seed):
        dim = out_info.shape[0] if corder else B_info.shape[0]
        fn = jax_kernel(kernel, launch_dims=[dim], num_outputs=1, in_out_argnames=['posts'])
        return fn(w_loc, w_scale, clen, B, seed, jnp.zeros(out_info.shape, out_info.dtype))

    return run


def _jitnmm_pallas_kernel_generator(
    B_info: jax.ShapeDtypeStruct,
    out_info: jax.ShapeDtypeStruct,
    corder: bool = True,
    **kwargs
):
    from jax.experimental import pallas as pl
    from jax.experimental.pallas.triton import atomic_add  # type: ignore[assignment]
    _PallasLFSRRNG = get_pallas_lfsr_rng_class()

    B_cols = B_info.shape[1]

    if corder:
        # Match jitn corder=True: vectorize over rows with identical RNG seeding.
        # Grid: (row_blocks, B_cols). Each kernel block processes one B column,
        # using vector RNG identical to jitn. Accumulates into 1D local array
        # to avoid 2D local arrays (dynamic_slice unsupported in Pallas Triton).
        out_rows = out_info.shape[0]
        row_block = generate_block_dim(out_rows, maximum=128)
        grid = (pl.cdiv(out_rows, row_block), B_cols)

        def kernel(w_loc_ref, w_scale_ref, clen_ref, B_ref, seed_ref, _, post_ref):
            k = B_ref.shape[0]
            w_loc0 = w_loc_ref[0]
            w_scale0 = w_scale_ref[0]
            clen0 = clen_ref[0]
            seed0 = seed_ref[0]
            i_row_block = pl.program_id(0)
            col_j = pl.program_id(1)  # scalar B column index

            # Row indices — VECTOR, matching jitn exactly
            i_rows = i_row_block * row_block + jnp.arange(row_block)
            i_row_mask = i_rows < out_rows
            safe_rows = jnp.where(i_row_mask, i_rows, 0)

            # Seed each (row, col_j) thread with an additional column stride.
            rng = _PallasLFSRRNG(seed0 + i_rows * k)
            i_cols = rng.random_integers(0, clen0)  # [row_block]
            i_col_mask = i_cols < k

            out = jnp.zeros(row_block, dtype=post_ref.dtype)

            def body(data):
                i_cols, i_col_mask, rng, out = data
                w = rng.normal(w_loc0, w_scale0)  # [row_block]
                safe_cols = jnp.where(i_col_mask, i_cols, 0)
                b_vals = B_ref[safe_cols, col_j]  # [row_block] vector gather
                out += jnp.where(i_col_mask & i_row_mask, w * b_vals, 0.)
                i_cols += rng.random_integers(1, clen0)
                return i_cols, i_cols < k, rng, out

            _, _, _, out = jax.lax.while_loop(
                lambda data: jnp.sum(data[1]) > 0,
                body,
                (i_cols, i_col_mask, rng, out)
            )
            atomic_add(post_ref, (safe_rows, col_j), out, mask=i_row_mask)

    else:
        # Match jitn corder=False: vectorize over k-dim columns with identical
        # RNG seeding. Grid: (k_blocks, B_cols). Each block processes one B column.
        k_dim = B_info.shape[0]
        k_block = generate_block_dim(k_dim, maximum=128)
        grid = (pl.cdiv(k_dim, k_block), B_cols)

        def kernel(w_loc_ref, w_scale_ref, clen_ref, B_ref, seed_ref, _, post_ref):
            m = post_ref.shape[0]
            w_loc0 = w_loc_ref[0]
            w_scale0 = w_scale_ref[0]
            clen0 = clen_ref[0]
            seed0 = seed_ref[0]
            i_k_block = pl.program_id(0)
            col_j = pl.program_id(1)  # scalar B column index

            i_ks = i_k_block * k_block + jnp.arange(k_block)
            i_k_mask = i_ks < k_dim
            safe_ks = jnp.where(i_k_mask, i_ks, 0)

            # Preload B values for this column (1D vector gather)
            b_vals = B_ref[safe_ks, col_j]  # [k_block]

            # Seed each (k, col_j) thread with an additional column stride.
            rng = _PallasLFSRRNG(seed0 + i_ks * m)
            i_rows = rng.random_integers(0, clen0)
            i_row_mask = i_rows < m

            def body(data):
                i_rows, i_row_mask, rng = data
                w = rng.normal(w_loc0, w_scale0)  # [k_block]
                vals = jnp.where(i_k_mask & i_row_mask, w * b_vals, 0.)
                safe_rows = jnp.where(i_row_mask, i_rows, 0)
                atomic_add(post_ref, (safe_rows, col_j), vals,
                           mask=i_k_mask & i_row_mask)
                i_rows += rng.random_integers(1, clen0)
                return i_rows, i_rows < m, rng

            jax.lax.while_loop(
                lambda data: jnp.sum(data[1]) > 0,
                body,
                (i_rows, i_row_mask, rng)
            )

    def run(w_loc, w_scale, clen, B, seed):
        fn = pl.pallas_call(
            kernel,
            grid=grid,
            input_output_aliases={5: 0},
            out_shape=kwargs['outs'],
            backend='triton',
        )
        placeholder = jnp.zeros(kwargs['outs'][0].shape, kwargs['outs'][0].dtype)
        return fn(w_loc, w_scale, clen, B, seed, placeholder)

    return run


def _jitnmm_jvp_wloc(w_dot, w_loc, w_scale, clen, B, seed, *, shape, transpose, corder, backend, **kwargs):
    return jitnmm_p_call(
        w_dot, w_scale, clen, B, seed, shape=shape, transpose=transpose, corder=corder, backend=backend
    )


def _jitnmm_jvp_wscale(w_dot, w_loc, w_scale, clen, B, seed, *, shape, transpose, corder, backend, **kwargs):
    return jitnmm_p_call(w_loc, w_dot, clen, B, seed, shape=shape, transpose=transpose, corder=corder, backend=backend)


def _jitnmm_jvp_B(B_dot, w_loc, w_scale, clen, B, seed, *, shape, transpose, corder, backend, **kwargs):
    return jitnmm_p_call(
        w_loc, w_scale, clen, B_dot, seed, shape=shape, transpose=transpose, corder=corder, backend=backend
    )


def _jitnmm_transpose_rules(ct, w_loc, w_scale, clen, B, seed, *, shape, transpose, corder, backend, **kwargs):
    assert not ad.is_undefined_primal(clen)
    assert not ad.is_undefined_primal(seed)

    ct = ct[0]
    if ad.is_undefined_primal(B):
        r = jitnmm_p_call(
            w_loc,
            w_scale,
            clen,
            ct,
            seed,
            shape=shape,
            transpose=not transpose,
            corder=not corder,
            backend=backend,
        )[0]
        return w_loc, w_scale, clen, r, seed
    elif ad.is_undefined_primal(w_loc):
        # M = (w_loc + w_scale * Z) * mask
        # d(M @ B)/d(w_loc) = mask @ B
        # d(loss)/d(w_loc) = sum((mask^T @ ct) * B)
        r = jitnmm_p_call(
            1., 0., clen, ct, seed,
            shape=shape, transpose=not transpose, corder=not corder,
            backend=backend,
        )[0]
        dw_loc = jnp.expand_dims(jnp.sum(r * B), axis=0)
        return dw_loc, w_scale, clen, B, seed
    elif ad.is_undefined_primal(w_scale):
        # M = (w_loc + w_scale * Z) * mask
        # d(M @ B)/d(w_scale) = (Z * mask) @ B
        # d(loss)/d(w_scale) = sum(((Z*mask)^T @ ct) * B)
        r = jitnmm_p_call(
            0., 1., clen, ct, seed,
            shape=shape, transpose=not transpose, corder=not corder,
            backend=backend,
        )[0]
        dw_scale = jnp.expand_dims(jnp.sum(r * B), axis=0)
        return w_loc, dw_scale, clen, B, seed
    else:
        raise NotImplementedError(
            'Transpose rules for jitc_matmat_normal not implemented for '
            'non-undefined primals.'
        )


def _batching_axis1(args, axis=1, **kwargs):
    assert args[3].ndim == 3, 'Batching axis 0 requires 3D input.'
    m, maybe_batch1, maybe_batch2 = args[3].shape
    B = args[3].reshape(m, maybe_batch1 * maybe_batch2)
    r = jitnmm_p_call(
        args[0],
        args[1],
        args[2],
        B,
        args[4],
        shape=kwargs['shape'],
        transpose=kwargs['transpose'],
        corder=kwargs['corder'],
        backend=kwargs.get('backend'),
    )
    r = jnp.reshape(r[0], [r[0].shape[0], maybe_batch1, maybe_batch2])
    return [r], [axis]


def _jitnmm_batching(args, axes, **kwargs):
    if tuple(axes) == (None, None, None, 0, None):
        assert args[3].ndim == 3, 'Batching axis 0 requires 3D input.'
        args = list(args)
        args[3] = jnp.transpose(args[3], (1, 0, 2))
        return _batching_axis1(args, **kwargs)

    elif tuple(axes) == (None, None, None, 1, None):
        return _batching_axis1(args, **kwargs)

    elif tuple(axes) == (None, None, None, 2, None):
        return _batching_axis1(args, axis=2, **kwargs)

    else:
        return general_batching_rule(jitnmm_p, args, axes, **kwargs)


def _jitnmm_benchmark_data(*, platform):
    n_pre, n_post, prob, dtype = 1000, 1000, 0.1, jnp.float32
    configs = []
    for transpose in (False, True):
        for corder in (True, False):
            w_loc = jnp.ones(1, dtype=dtype)
            w_scale = jnp.ones(1, dtype=dtype) * 0.1
            clen = jnp.atleast_1d(jnp.asarray(2.0 / prob, dtype=dtype))
            b_rows = n_post if not transpose else n_pre
            B = jnp.asarray(np.random.randn(b_rows, 10), dtype=dtype)
            seed = jnp.asarray(42, dtype=jnp.uint32)
            name = f"{'T' if transpose else 'NT'},{'corder' if corder else 'rorder'}"
            configs.append(
                BenchmarkConfig(
                    name,
                    (w_loc, w_scale, clen, B, seed),
                    {'shape': (n_pre, n_post), 'transpose': transpose, 'corder': corder}
                )
            )
    return configs


def jitnmm_p_call(
    w_loc, w_scale, clen, B, seed, *,
    shape: MatrixShape, transpose: bool, corder: bool,
    backend
):
    """Dispatch the JIT normal matrix-matrix multiply primitive.

    Parameters
    ----------
    w_loc : jax.Array
        Weight mean, shape ``(1,)``.
    w_scale : jax.Array
        Weight standard deviation, shape ``(1,)``.
    clen : jax.Array
        Connection length, shape ``(1,)``.
    B : jax.Array
        Right-hand 2-D matrix.
    seed : jax.Array
        RNG seed, shape ``(1,)``.
    shape : tuple of int
        Logical matrix shape ``(n_pre, n_post)``.
    transpose : bool
        Whether to use the transposed matrix.
    corder : bool
        Column-major iteration order.
    backend : str or None, optional
        Compute backend.

    Returns
    -------
    tuple of jax.Array
        Single-element tuple containing the output matrix.

    Raises
    ------
    AssertionError
        If ``shape`` is not length 2, or if ``B`` is not 2-D, or if
        ``w_loc``, ``w_scale``, ``clen``, ``seed`` do not have the
        expected shapes, or if the matrix dimensions are incompatible.

    See Also
    --------
    jitnmm : High-level wrapper with unit support.

    Notes
    -----
    The product is computed without materialising the full matrix:

        ``Y[i, c] = sum_{j} Normal(w_loc, w_scale) * Bernoulli(prob) * B[j, c]``

    Each weight is generated on the fly using a deterministic PRNG seeded
    by ``seed``.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._jit_normal.float import jitnmm_p_call
        >>> out = jitnmm_p_call(
        ...     jnp.array([0.0]), jnp.array([1.0]),
        ...     jnp.array([20.0]), jnp.ones((5, 3)), jnp.array([42]),
        ...     shape=(10, 5), transpose=False, corder=True)
    """
    w_loc = jnp.atleast_1d(w_loc)
    w_scale = jnp.atleast_1d(w_scale)
    clen = jnp.atleast_1d(clen)

    assert len(shape) == 2, "The matrix shape should be a tuple of two integers."
    assert B.ndim == 2, "The input matrix B should be a 2D array."
    assert seed.ndim == 1, "The seed should be a 1D array."
    assert w_loc.ndim == 1, "The weight should be a 1D array."
    assert w_scale.ndim == 1, "The weight should be a 1D array."
    assert clen.ndim == 1, "The clen should be a 1D array."
    assert w_loc.shape == (1,), "The weight should be a scalar."
    assert w_scale.shape == (1,), "The weight should be a scalar."
    assert clen.shape == (1,), "The clen should be a scalar."
    assert seed.shape == (1,), "The seed should be a scalar."
    if transpose:
        assert shape[0] == B.shape[0], f"The matrix shape and B shape do not match. {B.shape} @ {shape}"
    else:
        assert shape[1] == B.shape[0], f"The matrix shape and B shape do not match. {shape} @ {B.shape}"

    out_info = (
        jax.ShapeDtypeStruct([shape[1], B.shape[1]], w_loc.dtype)
        if transpose else
        jax.ShapeDtypeStruct([shape[0], B.shape[1]], w_loc.dtype)
    )

    return jitnmm_p(
        w_loc,
        w_scale,
        clen,
        B,
        seed,
        outs=[out_info],
        w_loc_info=jax.ShapeDtypeStruct(w_loc.shape, w_loc.dtype),
        w_scale_info=jax.ShapeDtypeStruct(w_scale.shape, w_scale.dtype),
        clen_info=jax.ShapeDtypeStruct(clen.shape, clen.dtype),
        B_info=jax.ShapeDtypeStruct(B.shape, B.dtype),
        seed_info=jax.ShapeDtypeStruct(seed.shape, seed.dtype),
        out_info=out_info,
        shape=shape,
        transpose=transpose,
        corder=corder,
        backend=backend,
    )


jitnmm_p = XLACustomKernel(
    'float_jitnmm',
    doc="""
Low-level XLA custom-kernel primitive for ``jitnmm``.

This ``XLACustomKernel`` instance dispatches the JIT normal connectivity matrix-matrix
multiplication with floating-point weights operation to registered backends
(``numba``, ``warp``, ``pallas``), using runtime shape/dtype metadata provided by
the high-level wrapper.

In this operation, the connectivity matrix has weights normally distributed with specified
mean and standard deviation, and the input matrix contains floating-point values. Each
column of the input is processed independently in a standard matrix-matrix product.

Beyond backend dispatch, the primitive stores JAX transformation bindings
(JVP, transpose, batching, and call registration) so the operation integrates
correctly with ``jit``, ``vmap``, and autodiff.

Available backends can be queried with ``jitnmm_p.available_backends(platform)``,
and the default backend can be configured with ``jitnmm_p.set_default(platform, backend)``.

See Also
--------
jitnmm : High-level user-facing function wrapper.
"""
)
jitnmm_p.def_numba_kernel(_jitnmm_numba_kernel_generator)
jitnmm_p.def_warp_kernel(_jitnmm_warp_kernel_generator)
jitnmm_p.def_pallas_kernel('gpu', _jitnmm_pallas_kernel_generator)
jitnmm_p.def_jvp_rule2(_jitnmm_jvp_wloc, _jitnmm_jvp_wscale, None, _jitnmm_jvp_B, None)
jitnmm_p.def_transpose_rule(_jitnmm_transpose_rules)
jitnmm_p.def_batching_rule(_jitnmm_batching)
jitnmm_p.def_tags('jit_normal', 'float')
jitnmm_p.def_benchmark_data(_jitnmm_benchmark_data)
